/*
 * dvb.c: driver for the Siemens Fujitsu DVB PCI card
 *
 * Copyright (C) 1999-2001 Ralph  Metzler 
 *                       & Marcus Metzler for convergence integrated media GmbH
 *
 * originally based on code by:
 *
 * Copyright (C) 1998,1999 Christian Theiss <mistert@rz.fh-augsburg.de>
 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 * 

 * The author can be reached at ralph@convergence.de, 

 * the project's page is at http://linuxtv.org/dvb/
 */

#define __KERNEL_SYSCALLS__
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kmod.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/timer.h>
#include <linux/unistd.h>
#include <linux/byteorder/swabb.h>
#include <linux/poll.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <stdarg.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/pci.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <asm/semaphore.h>
#include <linux/init.h>
#include <linux/vmalloc.h>

#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>


#include "tuner.h"
#include "dvb.h"

#define	DPRAM_BASE 0x4000
#include "dvb_dpram4k.h"
#include "dvb_comcode.h"
#include "dvb_types.h"

#include "saa7146_core.h"
#include "saa7146_v4l.h"
#include "saa7146_defs.h"

#ifdef MODULE
MODULE_PARM(debug,"i");
MODULE_PARM(fdebug,"i");
MODULE_PARM(vidmem,"l");
MODULE_PARM(vidlow,"l");
MODULE_PARM(vidmode,"i");
MODULE_PARM(init_chan,"i");
MODULE_PARM(pids_off,"i");
MODULE_PARM(readfirm,"i");
MODULE_PARM(adac,"i");
#endif

static inline int i2c_writereg(struct dvb_struct *dvb, u8 id, u8 reg, u8 val);
int secSetTone(struct dvb_struct *dvb, secToneMode mode);
int qpsk_tune(struct dvb_struct *dvb, struct qpskParameters *para);
int qam_tune(struct dvb_struct *dvb, struct qamParameters *para);
int secSetVoltage(struct dvb_struct *dvb, secVoltage voltage);
static inline u8 i2c_readreg(struct dvb_struct *dvb, u8 id, u8 reg);
static void InitFront(struct dvb_struct *dvb);
static int  outcom(struct dvb_struct *dvb, int type, int com, int num, ...);
static void SetMode(struct dvb_struct *dvb, int mode);

void twrite(u8 const *buf);
void init_p2t(p2t_t *p, dvb_demux_feed_t *feed);
long int find_pes_header(u8 const *buf, long int length, int *frags);
void pes_to_ts(u8 const *buf, long int length, u16 pid, p2t_t *p);
void p_to_t(u8 const *buf, long int length, u16 pid, u8 *counter, dvb_demux_feed_t *feed);
static void add_qpsk_event(qpsk_t *qpsk, struct qpskEvent *ev);
static void add_qam_event(qam_t *qam, struct qamEvent *ev);

static u32 vidmem	= 0;
static u32 vidlow	= 0;

static int debug = 0;
#define dprintk	if (debug) printk

static int fdebug = 0;
#define fprintk	if (fdebug) printk

static int vidmode=CVBS_RGB_OUT;
static int init_chan=0;
static int pids_off=1;
static int readfirm=0;
static int adac=DVB_ADAC_TI;

static int num_dvb = 0;

#define saacomm(x,y) dvb->saa->command(dvb->i2cbus, (x), (y))

struct dvb_struct dvbs[MAX_NUM_DVB];


/****************************************************************************
 * General helper functions
 ****************************************************************************/

/* taken from linux/drivers/sound/sound_firmware.c */
/* Maybe the kernel should offer this somewhere else, not just in soundcore */

static int errno;

static int do_firmread(const char *fn, char **fp)
{
        int fd;
	loff_t l;
	char *dp;
       
	fd = open(fn, 0, 0);
	if (fd == -1) {
                dprintk(KERN_INFO "Unable to load '%s'.\n", fn);
		return 0;
	}
	l = lseek(fd, 0L, 2);
	if (l <= 0 || l > 300*1024) {
	        printk(KERN_INFO "Firmware too large '%s'\n", fn);
		sys_close(fd);
		return 0;
	}
	lseek(fd, 0L, 0);
	*fp= dp = vmalloc(l);
	if (dp == NULL)	{
		printk(KERN_INFO "Out of memory loading '%s'.\n", fn);
		sys_close(fd);
		return 0;
	}
	if (read(fd, dp, l) != l) {
		printk(KERN_INFO "Failed to read '%s'.\n", fn);
		vfree(dp);
		sys_close(fd);
		return 0;
	}
	close(fd);
	*fp = dp;
	return (int) l;
}

static
int firmread(const char *fn, char **fp)
{
	int r;
	mm_segment_t fs = get_fs();
        
	set_fs(get_ds());
	r = do_firmread(fn, fp);
	set_fs(fs);
	return r;
}

static inline void ddelay(int i) 
{
        current->state=TASK_INTERRUPTIBLE;
        schedule_timeout((HZ*i)/100);
}


/****************************************************************************
 * GPIO and DEBI functions
 ****************************************************************************/

#define saaread(adr)            saa7146_read(saamem,(adr))
#define saawrite(dat,adr) 	saa7146_write(saamem,(adr),(dat))

inline static void 
setgpio(struct dvb_struct *dvb, int port, u32 data)
{
        void *saamem=dvb->saa->mem;
        u32 val;

        val=saaread(GPIO_CTRL);
        val&=~(0xff << (8*(port)));
        val|=(data)<<(8*(port));
        saawrite(val, GPIO_CTRL);
}

/* This DEBI code is based on the Stradis driver 
   by Nathan Laredo <laredo@gnu.org> */

static int debiwait_maxwait = 5000;

static int wait_for_debi_done(struct dvb_struct *dvb)
{
        struct saa7146 *saa=dvb->saa;
        void *saamem=saa->mem;
	int i;

	/* wait for registers to be programmed */
	for (i = 0; i < 100000; i++)
                if (saaread(MC2) & 2)
                        break;;
	/* wait for transfer to complete */
	for (i = 0; i < 500000 &&
	     (saaread(PSR) & SPCI_DEBI_S); i++)
		saaread(MC2);
	if (i > debiwait_maxwait)
		printk("wait-for-debi-done maxwait: %d\n",
			debiwait_maxwait = i);
	
	if (i == 500000)
		return -1;
	return 0;
}

static int debiwrite(struct dvb_struct *dvb, u32 config, 
                     int addr, u32 val, int count)
{
        struct saa7146 *saa=dvb->saa;
        void *saamem=saa->mem;
	u32 cmd;

	if (count <= 0 || count > 32764)
		return -1;
	if (wait_for_debi_done(dvb) < 0)
		return -1;
	saawrite(config, DEBI_CONFIG);
	if (count <= 4)		/* immediate transfer */
		saawrite(val, DEBI_AD);
	else			/* block transfer */
		saawrite(virt_to_bus(saa->debi), DEBI_AD);
	saawrite((cmd = (count << 17) | (addr & 0xffff)), DEBI_COMMAND);
	saawrite((2 << 16) | 2, MC2);
	return 0;
}

static u32 debiread(struct dvb_struct *dvb, u32 config, int addr, int count)
{
        struct saa7146 *saa=dvb->saa;
        void *saamem=saa->mem;
	u32 result = 0;

	if (count > 32764 || count <= 0)
		return 0;
	if (wait_for_debi_done(dvb) < 0)
		return 0;
	saawrite(virt_to_bus(saa->debi), DEBI_AD);
	saawrite((count << 17) | 0x10000 | (addr & 0xffff),
		 DEBI_COMMAND);
	saawrite(config, DEBI_CONFIG);
	saawrite((2 << 16) | 2, MC2);
	if (count > 4)	
		return count;
	wait_for_debi_done(dvb);
	result = saaread(DEBI_AD);
        result &= (0xffffffffUL >> ((4-count)*8));
	return result;
}

/* DEBI during interrupt */

static inline void 
iwdebi(struct dvb_struct *dvb, u32 config, int addr, u32 val, int count)
{
        if (count>4 && val)
                memcpy(dvb->saa->debi, (char *) val, count);
        debiwrite(dvb, config, addr, val, count);
}

static inline u32 
irdebi(struct dvb_struct *dvb, u32 config, int addr, u32 val, int count)
{
        u32 res;

        res=debiread(dvb, config, addr, count);
        if (count<=4) 
                memcpy(dvb->saa->debi, (char *) &res, count);
        return res;
}

/* DEBI outside interrupts, only for count<=4! */

static inline void 
wdebi(struct dvb_struct *dvb, u32 config, int addr, u32 val, int count)
{
        unsigned long flags;

        spin_lock_irqsave(&dvb->debilock, flags);
        debiwrite(dvb, config, addr, val, count);
        spin_unlock_irqrestore(&dvb->debilock, flags);
}

static inline u32 
rdebi(struct dvb_struct *dvb, u32 config, int addr, u32 val, int count)
{
        unsigned long flags;
        u32 res;

        spin_lock_irqsave(&dvb->debilock, flags);
        res=debiread(dvb, config, addr, count);
        spin_unlock_irqrestore(&dvb->debilock, flags);
        return res;
}


static inline char 
chtrans(char c)
{
        if (c<32 || c>126)
                c=0x20;
        return c;
}


/* handle mailbox registers of the dual ported RAM */

static inline void 
ARM_ResetMailBox(struct dvb_struct *dvb)
{
        unsigned long flags;

        spin_lock_irqsave(&dvb->debilock, flags);
        debiread(dvb, DEBINOSWAP, IRQ_RX, 2);
        //printk("dvb: IRQ_RX=%d\n", debiread(dvb, DEBINOSWAP, IRQ_RX, 2));
        debiwrite(dvb, DEBINOSWAP, IRQ_RX, 0, 2);
        spin_unlock_irqrestore(&dvb->debilock, flags);
}

static inline void 
ARM_ClearMailBox(struct dvb_struct *dvb)
{
        iwdebi(dvb, DEBINOSWAP, IRQ_RX, 0, 2);
}

static inline void 
ARM_ClearIrq(struct dvb_struct *dvb)
{
	irdebi(dvb, DEBINOSWAP, IRQ_RX, 0, 2);
}

static void 
reset_arm(struct dvb_struct *dvb)
{
        setgpio(dvb, RESET_LINE, GPIO_OUTLO);
        ddelay(1);
        setgpio(dvb, RESET_LINE, GPIO_OUTHI);
        
        printk("dvb: ARM RESET\n");
        ddelay(200);

        ARM_ClearIrq(dvb);
        InitFront(dvb);
}

static void 
arm_error(struct dvb_struct *dvb)
{
        dvb->arm_errors++;
        if (dvb->arm_errors<2)
                return;
        reset_arm(dvb);
        dvb->arm_errors=0;
}

static int
record_cb(pes2ts_t *p2t, u8 *buf, size_t len)
{
        dvb_demux_feed_t *dvbdmxfeed=(dvb_demux_feed_t *) p2t->priv;

        if (!(dvbdmxfeed->ts_type & TS_PACKET)) 
                return 0;
	if (buf[3]==0xe0)        // video PES do not have a length in TS
                buf[4]=buf[5]=0;
        if (dvbdmxfeed->ts_type & TS_PAYLOAD_ONLY) 
                return dvbdmxfeed->cb.ts(buf, len, 0, 0, 
                                         &dvbdmxfeed->feed.ts, DMX_OK); 
        else
                return pes2ts(p2t, buf, len);
}

static int 
pes2ts_cb(void *priv, unsigned char *data)
{
        dvb_demux_feed_t *dvbdmxfeed=(dvb_demux_feed_t *) priv;
        
        dvbdmxfeed->cb.ts(data, 188, 0, 0,
                          &dvbdmxfeed->feed.ts,
                          DMX_OK); 
        return 0;
}

static int 
AV_StartRecord(struct dvb_struct *dvb, int av,
               dvb_demux_feed_t *dvbdmxfeed)
{
        dvb_demux_t *dvbdmx=dvbdmxfeed->demux;
  
        if (dvb->playing||(dvb->rec_mode&av))
                return -EBUSY;
        outcom(dvb, COMTYPE_REC_PLAY, __Stop, 0);
        dvbdmx->recording=1;
        dvb->rec_mode|=av;

        switch (dvb->rec_mode) {
        case RP_AUDIO:
                pes2ts_init(&dvb->p2t[0], dvbdmx->pesfilter[0]->pid,
                            pes2ts_cb, (void *)dvbdmx->pesfilter[0]);
                outcom(dvb, COMTYPE_REC_PLAY, __Record, 2, AudioPES, 0);
                break;
        case RP_VIDEO:
                pes2ts_init(&dvb->p2t[1], dvbdmx->pesfilter[1]->pid,
                            pes2ts_cb, (void *)dvbdmx->pesfilter[1]);
                outcom(dvb, COMTYPE_REC_PLAY, __Record, 2, VideoPES, 0);
                break;
        case RP_AV:
                pes2ts_init(&dvb->p2t[0], dvbdmx->pesfilter[0]->pid,
                            pes2ts_cb, (void *)dvbdmx->pesfilter[0]);
                pes2ts_init(&dvb->p2t[1], dvbdmx->pesfilter[1]->pid,
                            pes2ts_cb, (void *)dvbdmx->pesfilter[1]);
                outcom(dvb, COMTYPE_REC_PLAY, __Record, 2, AV_PES, 0);
                break;
        }
        return 0;
}

static int 
AV_StartPlay(struct dvb_struct *dvb, int av)
{
        if (dvb->rec_mode)
                return -EBUSY;
        if (dvb->playing&av)
                return -EBUSY;

        outcom(dvb, COMTYPE_REC_PLAY, __Stop, 0);

        if (dvb->playing == RP_NONE) {
                reset_ipack(&dvb->ipack[0]);
                reset_ipack(&dvb->ipack[1]);
        }

        dvb->playing|=av;
        switch (dvb->playing) {
        case RP_AUDIO:
                outcom(dvb, COMTYPE_REC_PLAY, __Play, 2, AudioPES, 0);
                break;
        case RP_VIDEO:
                outcom(dvb, COMTYPE_REC_PLAY, __Play, 2, VideoPES, 0);
                dvb->sinfo=0;
                break;
        case RP_AV:
                dvb->sinfo=0;
                outcom(dvb, COMTYPE_REC_PLAY, __Play, 2, AV_PES, 0);
                break;
        }
        return dvb->playing;
}

static void 
AV_Stop(struct dvb_struct *dvb, int av)
{
        if (!(dvb->playing&av) && !(dvb->rec_mode&av))
                return;

        outcom(dvb, COMTYPE_REC_PLAY, __Stop, 0);
        if (dvb->playing) {
                dvb->playing&=~av;
                switch (dvb->playing) {
                case RP_AUDIO:
                        outcom(dvb, COMTYPE_REC_PLAY, __Play, 2, AudioPES, 0);
                        break;
                case RP_VIDEO:
                        outcom(dvb, COMTYPE_REC_PLAY, __Play, 2, VideoPES, 0);
                        break;
                case RP_NONE:
                        SetMode(dvb, dvb->vidmode);
                        break;
                }
        } else {
                dvb->rec_mode&=~av;
                switch (dvb->rec_mode) {
                case RP_AUDIO:
                        outcom(dvb, COMTYPE_REC_PLAY, __Record, 2, AudioPES, 0);
                        break;
                case RP_VIDEO:
                        outcom(dvb, COMTYPE_REC_PLAY, __Record, 2, VideoPES, 0);
                        break;
                case RP_NONE:
                        break;
                }
        }
}

/****************************************************************************
 * Buffer handling
 ****************************************************************************/

static inline void 
dvb_buf_flush(struct dvb_buf *dbuf)
{
        spin_lock_irq(&dbuf->lock);
        dbuf->head=dbuf->tail=0;
        spin_unlock_irq(&dbuf->lock);
        wake_up_interruptible(&dbuf->queue);
}

static inline void 
dvb_buf_init(struct dvb_buf *dbuf, u8 *data, int len)
{
        dbuf->head=dbuf->tail=0;
        dbuf->buf=data;
        dbuf->length=len;
        init_waitqueue_head(&dbuf->queue);
        spin_lock_init (&(dbuf->lock));
        dbuf->lock=SPIN_LOCK_UNLOCKED;
        sema_init(&(dbuf->sema), 1);
}

static inline 
int dvb_buf_empty(dvb_buf_t *buf)
{
        return (buf->head==buf->tail);
}

static inline 
int dvb_buf_free(dvb_buf_t *buf)
{
        int free;
  
        free=buf->head - buf->tail;
        if (free<=0)
                free+=buf->length;
        return free;
}

static inline 
int dvb_buf_avail(dvb_buf_t *buf)
{
        int avail;
  
        avail=buf->tail - buf->head;
        if (avail<0)
                avail+=buf->length;
        return avail;
}

#if 0
static void
dvb_buf_block(dvb_buf_t *dbuf, unsigned long count)
{
        if (dvb_buf_free(dbuf)>=count)
                return;
        while (!wait_event_interruptible(dbuf->queue,
                                         (dvb_buf_free(dbuf)>=count)));
}
#endif

static long 
dvb_buf_write(dvb_buf_t *dbuf, 
              const char *buf, unsigned long count, 
              int nonblock, int usermem)
{
        unsigned long todo = count;
        int free, split;
    
        while (todo > 0) {
                if (dvb_buf_free(dbuf)<=2) {
                        if (nonblock)
                                return count-todo;
                        if (wait_event_interruptible(dbuf->queue,
                                                     (dvb_buf_free(dbuf)>2)))
                        	return count-todo;
                }   
                free = dbuf->head - dbuf->tail;
                split=dbuf->length;
                if (free<=0) {
                        free+=dbuf->length;
                        split-=dbuf->tail;
                }
                free-=2; 

                if (free > todo)
                        free = todo;
                        
                if (split < free) {
                        if (!usermem) 
                                memcpy(dbuf->buf+dbuf->tail, buf, split);
                        else
                                if (copy_from_user(dbuf->buf+dbuf->tail,
                                                   buf, split)) 
                                        return -EFAULT;
                        buf += split;
                        todo -= split;
                        free -= split;
                        dbuf->tail = 0;
                }
                if (!usermem) 
                        memcpy(dbuf->buf+dbuf->tail, buf, free);
                else
                        if (copy_from_user(dbuf->buf+dbuf->tail, buf, free)) 
                                return -EFAULT;
                dbuf->tail = (dbuf->tail + free)%dbuf->length;
                todo -= free;
                buf += free;
        }

	return count-todo;
}

#if 0
static void
dvb_buf_put(struct dvb_buf *db, u8 *buf, int len)
{
        int split, fsize;

        fsize=db->head - db->tail;
        if (fsize <= 0) {
                fsize+=db->length;
                split=db->length-db->tail;
        } else 
                split=0;
        if (len>=fsize) {
                dprintk("buffer overflow\n");
                return;
        }
        if (split>=len)
                split=0;
        if (split) {
                memcpy(db->buf + db->tail, buf, split);
                len-=split;
                db->tail=0;
        }
        memcpy(db->buf + db->tail, split + buf, len);
        db->tail=(db->tail+len)%db->length;
}
#endif


/****************************************************************************
 * TT budget / WinTV Nova
 ****************************************************************************/

static int 
TTBStop(struct dvb_struct *dvb)
{
        if (--dvb->feeding)
                return dvb->feeding;
        saa7146_write(dvb->saa->mem, MC1, MASK_20); // DMA3 off
        saa7146_write(dvb->saa->mem, MC1, MASK_28); // RPS0 off
        saa7146_write(dvb->saa->mem, IER, 
                      saa7146_read(dvb->saa->mem, IER) & ~MASK_10 );
        saa7146_write(dvb->saa->mem, IER, 
                      saa7146_read(dvb->saa->mem, IER)& ~MASK_07);
        return 0;
}

#define TS_WIDTH  (4*188)
#define TS_HEIGHT (1024/4)
static int
TTBStart(struct dvb_struct *dvb)
{
        struct saa7146 *saa=dvb->saa;

        //printk ("function : %s\n", __FUNCTION__);
        if (dvb->feeding++)
                return dvb->feeding;

      	saa7146_write(saa->mem, MC1, MASK_20); // DMA3 off

        memset(saa->grabbing, 0x00, TS_HEIGHT*TS_WIDTH);

        saa7146_write(saa->mem, PCI_BT_V1, 0x001c0000);

        dvb->tsf=0;
        saa7146_write(dvb->saa->mem, DD1_INIT, 0x02000680);
        saa7146_write(dvb->saa->mem, MC2, 
                      (MASK_09 | MASK_25 | MASK_10 | MASK_26));

        saa7146_write(saa->mem, BRS_CTRL, 0x60000000);	
      	saa7146_write(saa->mem, MC2, (MASK_08 | MASK_24));
        ddelay(1);

        saa7146_write(saa->mem, BASE_ODD3, 0);
        saa7146_write(saa->mem, BASE_EVEN3, TS_WIDTH*TS_HEIGHT/2);
        saa7146_write(saa->mem, PROT_ADDR3, TS_WIDTH*TS_HEIGHT);	
        saa7146_write(saa->mem, BASE_PAGE3, virt_to_bus(saa->page_table[0])|ME1|0xb0);
        saa7146_write(saa->mem, PITCH3, TS_WIDTH);	

        saa7146_write(saa->mem, NUM_LINE_BYTE3, ((TS_HEIGHT/2)<<16)|TS_WIDTH);
      	saa7146_write(saa->mem, MC2, (MASK_04 | MASK_20));

        // VPE
        saa7146_write(dvb->saa->mem, IER, saa7146_read(dvb->saa->mem, IER)|MASK_10);

     	saa7146_write(saa->mem, MC1, (MASK_04 | MASK_20)); // DMA3 on

        // FIDB
        saa7146_write(dvb->saa->mem, IER, saa7146_read(dvb->saa->mem, IER)|MASK_07);
        return dvb->feeding;
}


/****************************************************************************
 * IRQ handling
 ****************************************************************************/

void IR_handle(struct dvb_struct *dvb, u32 ircom)
{
        printk("dvb: ircommand = %08x\n", ircom);
}

void CI_handle(struct dvb_struct *dvb, u8 *data, u16 len) 
{
        //CI_out(dvb, data, len);

        if (len<3)
                return;
        switch (data[0]) {
        case CI_MSG_CI_INFO:
                if (data[2]!=1 && data[2]!=2)
                        break;
                switch (data[1]) {
                case 0:
                        dvb->ci_slot[data[2]-1]=0;
                        break;
                case 1:
                        dvb->ci_slot[data[2]-1]|=CI_MODULE_PRESENT;
                        break;
                case 2:
                        dvb->ci_slot[data[2]-1]|=CI_MODULE_OK;
                        break;
                }
                break;
        case CI_SWITCH_PRG_REPLY:
                dvb->ci_stat=data[1];
                break;
        default:
                break;
        }

}

static inline int
DvbDmxFilterCallback(__u8 * buffer1, size_t buffer1_len,
                     __u8 * buffer2, size_t buffer2_len,
                     dvb_demux_filter_t *dvbdmxfilter,
                     dmx_success_t success,
                     struct dvb_struct *dvb)
{
        switch(dvbdmxfilter->type) {
        case DMX_TYPE_SEC:
                if ((((buffer1[1]<<8)|buffer1[2])&0xfff)+3!=buffer1_len)
                        return 0;
                return dvbdmxfilter->feed->cb.sec(buffer1, buffer1_len,
						  buffer2, buffer2_len,
						  &dvbdmxfilter->filter,
						  DMX_OK); 
        case DMX_TYPE_TS:
                if (!(dvbdmxfilter->feed->ts_type & TS_PACKET)) 
                        return 0;
                if (dvbdmxfilter->feed->ts_type & TS_PAYLOAD_ONLY) 
                        return dvbdmxfilter->feed->cb.ts(buffer1, buffer1_len,
                                                         buffer2, buffer2_len,
                                                         &dvbdmxfilter->feed->feed.ts,
                                                         DMX_OK); 
                else
                        pes_to_ts(buffer1, buffer1_len, 
                                  dvbdmxfilter->feed->pid, 
                                  &dvb->p2t_filter[dvbdmxfilter->index]);
	default:
	        return 0;
        }
}


u8 pshead[0x26] = {
        0x00, 0x00, 0x01, 0xba, 0x5f, 0xff, 0xfe, 0xe6, 
        0xc4, 0x01, 0x01, 0x89, 0xc3, 0xf8, 0x00, 0x00,
        0x01, 0xbb, 0x00, 0x12, 0x80, 0xc4, 0xe1, 0x00,
        0xe1, 0xff, 0xb9, 0xe0, 0xe8, 0xb8, 0xc0, 0x20,
        0xbd, 0xe0, 0x44, 0xbf, 0xe0, 0x02,
};								


static void vpeirq(struct saa7146* saa, void *data)
{
        struct dvb_struct *dvb=(struct dvb_struct *) data;
  
        if (dvb->card_type!=DVB_CARD_TT_BUDGET)
                return;
        //printk("vpeirq %08x\n", saa7146_read(dvb->saa->mem, PCI_VDP3));
}

static void fidbirq(struct saa7146* saa, void *data)
{
        struct dvb_struct *dvb=(struct dvb_struct *) data;
        u8 *mem;
  
        if (dvb->card_type!=DVB_CARD_TT_BUDGET)
                return;
        
        mem=(dvb->tsf ? TS_HEIGHT*TS_WIDTH/2 :0)+(u8 *)dvb->saa->grabbing;

        dvb->tsf^=1;
        saa7146_write(dvb->saa->mem, DD1_INIT, 0x02000600|(dvb->tsf ? 0x40:0x80));
        saa7146_write(dvb->saa->mem, MC2, 
                      (MASK_09 | MASK_25 | MASK_10 | MASK_26));

        // FIXME: use bottom half or tasklet
        if (dvb->feeding && mem[0]==0x47)
                DvbDmxSWFilterPackets(&dvb->demux, mem, 512);
}

static void debiirq(struct saa7146* saa, void *data)
{
        struct dvb_struct *dvb=(struct dvb_struct *) data;
        int type=dvb->debitype;
        int handle=(type>>8)&0x1f;
        
        dprintk("DEBI irq type %d\n", type);        
//#define DEBUG_TIMING
#ifdef DEBUG_TIMING
        {
                struct timeval tv;
                do_gettimeofday(&tv);
                printk("debi%d:   %d.%d\n", type, tv.tv_sec, tv.tv_usec);
        }
#endif
        saa7146_write(dvb->saa->mem, IER, 
                      saa7146_read(dvb->saa->mem, IER) & ~MASK_19 );
        saa7146_write(dvb->saa->mem, ISR, MASK_19 );

        if (type==-1) {
                printk("DEBI irq oops\n");
                ARM_ClearMailBox(dvb);
                ARM_ClearIrq(dvb);
                return;
        }
        dvb->debitype=-1;

        switch (type&0xff) {

        case DATA_PES_RECORD:
                if (dvb->demux.recording) 
                        record_cb(&dvb->p2t[handle], 
                                  (u8 *)dvb->saa->debi,
                                  dvb->debilen);
                spin_lock(&dvb->debilock);
                iwdebi(dvb, DEBINOSWAP, RX_BUFF, 0, 2);
                ARM_ClearMailBox(dvb);
                spin_unlock(&dvb->debilock);
                return;

        case DATA_IPMPE:
        case DATA_FSECTION:
        case DATA_PIPING:
                if (dvb->handle2filter[handle]) 
                        DvbDmxFilterCallback((u8 *)dvb->saa->debi, dvb->debilen, 0, 0, 
                                             dvb->handle2filter[handle], DMX_OK, dvb); 
#if 0
                {
                        u8 *p=(u8 *)dvb->saa->debi;
                        printk("%02d: %02x %02x %02x %02x %02x %02x  %04x\n", 
                               handle,
                               p[0], p[1], p[2], p[3], p[4], p[5], dvb->debilen);
                }
#endif
                spin_lock(&dvb->debilock);
                iwdebi(dvb, DEBINOSWAP, RX_BUFF, 0, 2);
                ARM_ClearMailBox(dvb);
                spin_unlock(&dvb->debilock);
                return;

        case DATA_COMMON_INTERFACE:
                CI_handle(dvb, (u8 *)dvb->saa->debi, dvb->debilen);
#if 1
        {
                int i;

                printk("dvb%d: ", dvb->num);
                printk("%02x ", *(u8 *)dvb->saa->debi);
                printk("%02x ", *(1+(u8 *)dvb->saa->debi));
                for (i=2; i<dvb->debilen; i++)
                  printk("%02x ", (*(i+(char *)dvb->saa->debi)));
                for (i=2; i<dvb->debilen; i++)
                  printk("%c", chtrans(*(i+(char *)dvb->saa->debi)));

                printk("\n");
        }
#endif
                spin_lock(&dvb->debilock);
                iwdebi(dvb, DEBINOSWAP, RX_BUFF, 0, 2);
                ARM_ClearMailBox(dvb);
                spin_unlock(&dvb->debilock);
                return;

        case DATA_DEBUG_MESSAGE:
                ((s8*)dvb->saa->debi)[Reserved_SIZE-1]=0;
                printk("%s\n", (s8 *)dvb->saa->debi);
                spin_lock(&dvb->debilock);
                iwdebi(dvb, DEBINOSWAP, RX_BUFF, 0, 2);
                ARM_ClearMailBox(dvb);
                spin_unlock(&dvb->debilock);
                return;

        case DATA_MPEG_PLAY:
        case DATA_BMP_LOAD:
                spin_lock(&dvb->debilock);
                iwdebi(dvb, DEBINOSWAP, TX_BUFF, 0, 2);
                ARM_ClearMailBox(dvb);
                spin_unlock(&dvb->debilock);
                return;
        default:
                break;
        }
        spin_lock(&dvb->debilock);
        ARM_ClearMailBox(dvb);
        spin_unlock(&dvb->debilock);
}

static int
pes_play(void *dest, dvb_buf_t *buf, int dlen)
{
        int len, split=0;
        u32 sync;
        u16 blen;

        /*
        if (!dlen) {
                wake_up_interruptible(&buf->queue);
                return -1;
        }
        */
        while (1) {
                if ((len=dvb_buf_avail(buf)) < 6)
                        return -1;
                sync=(buf->buf[buf->head])<<24;
                sync|=(buf->buf[(buf->head+1)%buf->length]<<16);
                sync|=(buf->buf[(buf->head+2)%buf->length]<<8);
                sync|=buf->buf[(buf->head+3)%buf->length];
                
                if (((sync&~0x1f)==0x000001e0) ||
                    ((sync&~0x1f)==0x000001c0) ||
                    (sync==0x000001bd))
                        break;
                buf->head=(buf->head+1)%buf->length;
        }
        blen=(buf->buf[(buf->head+4)%buf->length]<<8);
        blen|=buf->buf[(buf->head+5)%buf->length];
        blen+=6;
        if (len<blen || blen > dlen) {
                dprintk("buffer empty\n");
                return -1;
        }
        if (blen>2048) {
                buf->head=(buf->head+blen)%buf->length;
                dprintk("packet too large\n");
                return -1;
        }
        len=blen;
        if (buf->head + len > buf->length)
                split=buf->length-buf->head;
        if (split>0) {
                memcpy(dest, buf->buf+buf->head, split);
                buf->head=0;
                len-=split;
        }
        memcpy(split + dest, 
               buf->buf + buf->head, len);
        buf->head = (buf->head +len)%buf->length;
        
        wake_up_interruptible(&buf->queue);
        return blen;
}

static void gpioirq(struct saa7146* saa, void *data)
{
        struct dvb_struct *dvb=(struct dvb_struct *) data;
        u32 rxbuf, txbuf;
        int len;
        
        dprintk("GPIO0 irq\n");        

        if (dvb->debitype !=-1)
                printk("GPIO0 irq oops\n");
       
        spin_lock(&dvb->debilock);

	ARM_ClearIrq(dvb);

        saa7146_write(dvb->saa->mem, IER, 
                      saa7146_read(dvb->saa->mem, IER) & ~MASK_19 );
        saa7146_write(dvb->saa->mem, ISR, MASK_19 );

        dvb->debitype = irdebi(dvb, DEBINOSWAP, IRQ_STATE, 0, 2);
        dvb->debilen  = irdebi(dvb, DEBINOSWAP, IRQ_STATE_EXT, 0, 2);
        dvb->debibuf  = 0;
        rxbuf=irdebi(dvb, DEBINOSWAP, RX_BUFF, 0, 2);
        txbuf=irdebi(dvb, DEBINOSWAP, TX_BUFF, 0, 2);

        dprintk("GPIO0 irq %d %d\n", dvb->debitype, dvb->debilen);

#ifdef DEBUG_TIMING
                {
                        struct timeval tv;
                        do_gettimeofday(&tv);
                        printk("gpio%d: %d.%d\n", dvb->debitype, tv.tv_sec, tv.tv_usec);
                        //udelay(300);
                        //wait_for_debi_done(dvb);
                        //do_gettimeofday(&tv);
                        //printk("gpio: %d.%d\n", tv.tv_sec, tv.tv_usec);
                }
#endif
        switch (dvb->debitype&0xff) {

        case DATA_MPEG_PLAY:
                if (!dvb->playing) {
                        iwdebi(dvb, DEBINOSWAP, IRQ_STATE_EXT, 0, 2);
                        iwdebi(dvb, DEBINOSWAP, TX_LEN, 0, 2);
                        iwdebi(dvb, DEBINOSWAP, TX_BUFF, 0, 2);
                        break;
                }
                len=0;

                if (dvb->debitype&0x100) {
                        spin_lock(&dvb->aout.lock);
                        len=pes_play(dvb->saa->debi, &dvb->aout, 2048);
                        spin_unlock(&dvb->aout.lock);
                        if (dvb->trickmode!=TRICK_NONE)
                                len=0;
                }
                if (len<=0 && (dvb->debitype&0x200) && 
                    dvb->videostate.playState!=VIDEO_FREEZED) {
                        spin_lock(&dvb->avout.lock);
                        len=pes_play(dvb->saa->debi, &dvb->avout, 2048);
                        spin_unlock(&dvb->avout.lock);
                }

                if (len<=0) {
                        iwdebi(dvb, DEBINOSWAP, IRQ_STATE_EXT, 0, 2);
                        iwdebi(dvb, DEBINOSWAP, TX_LEN, 0, 2);
                        iwdebi(dvb, DEBINOSWAP, TX_BUFF, 0, 2);
                        break;
                } 
                iwdebi(dvb, DEBINOSWAP, TX_LEN, len, 2);
                iwdebi(dvb, DEBINOSWAP, IRQ_STATE_EXT, len, 2);
                wait_for_debi_done(dvb);
                saa7146_write(dvb->saa->mem, IER, 
                              saa7146_read(dvb->saa->mem, IER) | MASK_19 );

                iwdebi(dvb, DEBISWAB, DPRAM_BASE+txbuf, 0, (len+3)&~3);
                spin_unlock(&dvb->debilock);
                return;

        case DATA_TS_PLAY:
                break;

        case DATA_PES_PLAY:
                break;

        case DATA_BMP_LOAD:
                len=dvb->debilen;
                if (!len) {
                        dvb->bmp_state=BMP_LOADED;
                        iwdebi(dvb, DEBINOSWAP, IRQ_STATE_EXT, 0, 2);
                        iwdebi(dvb, DEBINOSWAP, TX_LEN, 0, 2);
                        iwdebi(dvb, DEBINOSWAP, TX_BUFF, 0, 2);
                        wake_up_interruptible(&dvb->bmpq);
                        break;
                }
                if (len>dvb->bmplen)
                        len=dvb->bmplen;
                if (len>2*1024)
                        len=2*1024;
                iwdebi(dvb, DEBINOSWAP, TX_LEN, len, 2);
                iwdebi(dvb, DEBINOSWAP, IRQ_STATE_EXT, len, 2);
                memcpy(dvb->saa->debi, dvb->bmpbuf+dvb->bmpp, len);
                dvb->bmpp+=len;
                dvb->bmplen-=len;
                wait_for_debi_done(dvb);
                saa7146_write(dvb->saa->mem, IER, 
                              saa7146_read(dvb->saa->mem, IER) | MASK_19 );
                iwdebi(dvb, DEBISWAB, DPRAM_BASE+txbuf, 0, (len+3)&~3);
                spin_unlock(&dvb->debilock);
                return;

        case DATA_PES_RECORD:
                len=(dvb->debilen+3)&(~3);
                saa7146_write(dvb->saa->mem, IER, 
                              saa7146_read(dvb->saa->mem, IER) | MASK_19);
                irdebi(dvb, DEBISWAB, DPRAM_BASE+rxbuf, 0, len);
                spin_unlock(&dvb->debilock);
                return;

        case DATA_COMMON_INTERFACE:
        case DATA_FSECTION:
        case DATA_IPMPE:
        case DATA_PIPING:
                len=(dvb->debilen+3)&(~3);

                //printk("PIPING len=%08x\n", len);
                if (!len || len>4*1024) {
                        iwdebi(dvb, DEBINOSWAP, RX_BUFF, 0, 2);
                        break;
                }       

                saa7146_write(dvb->saa->mem, IER, 
                              saa7146_read(dvb->saa->mem, IER) | MASK_19);
                irdebi(dvb, DEBISWAB, DPRAM_BASE+rxbuf, 0, len);
                spin_unlock(&dvb->debilock);
                return;

        case DATA_DEBUG_MESSAGE:
                len=(dvb->debilen+3)&(~3);
                if (!len || len>0xff) {
                        iwdebi(dvb, DEBINOSWAP, RX_BUFF, 0, 2);
                        break;
                }
                saa7146_write(dvb->saa->mem, IER, 
                              saa7146_read(dvb->saa->mem, IER) | MASK_19);
                irdebi(dvb, DEBISWAB, Reserved, 0, len);
                spin_unlock(&dvb->debilock);
                return;

        case DATA_IRCOMMAND: 
                IR_handle(dvb,swahw32(irdebi(dvb, DEBINOSWAP, Reserved, 0, 4)));
                iwdebi(dvb, DEBINOSWAP, RX_BUFF, 0, 2);
                break;
        default:
                printk("gpioirq unknown type=%d len=%d\n", 
                       dvb->debitype, dvb->debilen);
                break;
        }      
        ARM_ClearMailBox(dvb);
        dvb->debitype=-1;
        spin_unlock(&dvb->debilock);
}



/****************************************************************************
 * DEBI command polling 
 ****************************************************************************/


static int OutCommand(struct dvb_struct *dvb, u16* buf, int length)
{
        int i;
        u32	start;

        start = jiffies;
        while ( rdebi(dvb, DEBINOSWAP, COMMAND, 0, 2 ) )
        {
                ddelay(1);
                if ((jiffies - start) > ARM_WAIT_FREE) {
                        printk("outcommand error 1\n");
                        arm_error(dvb);
                        return -1;
                }
        }

#ifndef _NOHANDSHAKE
        start = jiffies;
        while ( rdebi(dvb, DEBINOSWAP, HANDSHAKE_REG, 0, 2 ) ) 
        {
                ddelay(1);
                if ((jiffies - start) > ARM_WAIT_SHAKE) {
                        printk("outcommand error 2\n");
                        arm_error(dvb);
                        return -1;
                }
        }
#endif

        start = jiffies;
        while ( rdebi(dvb, DEBINOSWAP, MSGSTATE, 0, 2) & OSDQFull )
        {
                ddelay(1);
                if ((jiffies - start) > ARM_WAIT_OSD)
                {
                        printk("outcommand error 3\n");
                        arm_error(dvb);
                        return -1;
                }
        }
        for (i=2; i<length; i++)
                wdebi(dvb, DEBINOSWAP, COMMAND + 2*i, (u32) buf[i], 2);

        if (length)
                wdebi(dvb, DEBINOSWAP, COMMAND + 2, (u32) buf[1], 2);
        else
                wdebi(dvb, DEBINOSWAP, COMMAND + 2, 0, 2);
        wdebi(dvb, DEBINOSWAP, COMMAND, (u32) buf[0], 2);
        return 0;
}

inline static int 
SOutCommand(struct dvb_struct *dvb, u16* buf, int length)
{
        int ret;
        
        down_interruptible(&dvb->dcomlock);
        ret=OutCommand(dvb, buf, length);
        up(&dvb->dcomlock);
        return ret;
}


static int outcom(struct dvb_struct *dvb, int type, int com, int num, ...)
{
	va_list args;
        u16 buf[num+2];
        int i;

        buf[0]=(( type << 8 ) + com); 
        buf[1]=num;

        if (num) {
                va_start(args, num);
                for (i=0; i<num; i++)
                        buf[i+2]=va_arg(args, u32);
                va_end(args);
        }

        return SOutCommand(dvb, buf, num+2);
}

int SendCICommand(struct dvb_struct *dvb, u8 subcom, u8 *Params, u8 ParamLen)
{
        int i;
        u16 CommandBuffer[18] = { ((COMTYPE_COMMON_IF << 8) + subcom),
                                  16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
        
	for(i=0; (i<ParamLen)&&(i<32); i++)	
	{
		if(i%2 == 0)
			CommandBuffer[(i/2)+2] = (u16)(Params[i]) << 8;
		else
			CommandBuffer[(i/2)+2] |= Params[i];
	}

        return SOutCommand(dvb, CommandBuffer, 18);
}


static int CommandRequest(struct dvb_struct *dvb, u16 *Buff, int length, u16 *buf, int n)
{
        s16 i;
        u32 start;
       
        down_interruptible(&dvb->dcomlock);
        OutCommand(dvb, Buff, length);
        
        start = jiffies;
        while ( rdebi(dvb, DEBINOSWAP, COMMAND, 0, 2) ) 
        {
#ifdef _NOHANDSHAKE
                ddelay(1);
#endif
                if ((jiffies - start) > ARM_WAIT_FREE) {
                        printk("commandrequest error\n");
                        up(&dvb->dcomlock);
                        arm_error(dvb);
                        return -1;
                }
        }
        
#ifndef _NOHANDSHAKE
        start = jiffies;
        while ( rdebi(dvb, DEBINOSWAP, HANDSHAKE_REG, 0, 2 ) ) {
                ddelay(1);
                if ((jiffies - start) > ARM_WAIT_SHAKE) {
                        printk("commandrequest error\n");
                        up(&dvb->dcomlock);
                        arm_error(dvb);
                        return -1;
                }
        }
#endif

        for (i=0; i<n; i++)
                buf[i] = rdebi(dvb, DEBINOSWAP, COM_BUFF + 2*i, 0, 2);
        
	up(&dvb->dcomlock);
        return 0;
}


static inline int 
RequestParameter(struct dvb_struct *dvb, u16 tag, u16* Buff, s16 length)
{
        return CommandRequest(dvb, &tag, 0, Buff, length);
}

#if 0
static int
GetDecoderStatus(struct dvb_struct *dvb, PVIDEOSTATE vstate)
{
        if (RequestParameter(dvb, (COMTYPE_REQUEST << 8) + VideoState1, 
                             (u16*) vstate, 16)<0)
                return 0;
        if (RequestParameter(dvb, (COMTYPE_REQUEST << 8) + VideoState2, 
                             16+(u16*) vstate, 16)<0)
                return 0;
        if (RequestParameter(dvb, (COMTYPE_REQUEST << 8) + VideoState3, 
                             32+(u16*) vstate, 16)<0)
                return 0;

        return 0;
}
#endif



/****************************************************************************
 * Firmware commands 
 ****************************************************************************/


inline static int SendDAC(struct dvb_struct *dvb, u8 addr, u8 data)
{
        return outcom(dvb, COMTYPE_AUDIODAC, AudioDAC, 2, addr, data);
}

int static SetVolume(struct dvb_struct *dvb, int volleft, int volright)
{
        int err;
        
        if (dvb->dvbtype!=DVB_S)
                return -EOPNOTSUPP;

        switch (dvb->adac_type) {
        default:
        case DVB_ADAC_TI:
                volleft=(volleft*256)/946;
                volright=(volright*256)/946;
                if (volleft > 0x45)
                        volleft=0x45;
                if (volright > 0x45)
                        volright=0x45;
                err=SendDAC(dvb, 3, 0x80 + volleft);
                if (err)
                        return err;
                return SendDAC(dvb, 4, volright);
                
        case DVB_ADAC_CRYSTAL:
                volleft=127-(volleft/2);
                volright=127-(volright/2);
                i2c_writereg(dvb, 0x20, 0x00, 0x00);
                i2c_writereg(dvb, 0x20, 0x01, 0x08);
                i2c_writereg(dvb, 0x20, 0x02, 0x09);
                i2c_writereg(dvb, 0x20, 0x03, volleft);
                i2c_writereg(dvb, 0x20, 0x04, volright);
                return 0;
        }
}

#ifdef USE_OSD

inline static int ResetBlend(struct dvb_struct *dvb, u8 windownr)
{
        return outcom(dvb, COMTYPE_OSD, SetNonBlend, 1, windownr);
}

inline static int SetColorBlend(struct dvb_struct *dvb, u8 windownr)
{
        return outcom(dvb, COMTYPE_OSD, SetCBlend, 1, windownr); 
}

inline static int SetWindowBlend(struct dvb_struct *dvb, u8 windownr, u8 blending)
{
        return outcom(dvb, COMTYPE_OSD, SetWBlend, 2, windownr, blending); 
}

inline static int SetBlend_(struct dvb_struct *dvb, u8 windownr,
                     OSDPALTYPE colordepth, u16 index, u8 blending)
{
        return outcom(dvb, COMTYPE_OSD, SetBlend, 4,
                      windownr, colordepth, index, blending);
} 

inline static int SetColor_(struct dvb_struct *dvb, u8 windownr,
                     OSDPALTYPE colordepth, u16 index, u16 colorhi, u16 colorlo)
{
        return outcom(dvb, COMTYPE_OSD, SetColor, 5,
                      windownr, colordepth, index, colorhi, colorlo);
} 

inline static int BringToTop(struct dvb_struct *dvb, u8 windownr)
{
        return outcom(dvb, COMTYPE_OSD, WTop, 1, windownr);
} 

inline static int SetFont(struct dvb_struct *dvb, u8 windownr, u8 fontsize,
                   u16 colorfg, u16 colorbg)
{
        return outcom(dvb, COMTYPE_OSD, Set_Font, 4,
                      windownr, fontsize, colorfg, colorbg);
} 

static int FlushText(struct dvb_struct *dvb)
{
        u32 start;

        down_interruptible(&dvb->dcomlock);
        start = jiffies;
        while ( rdebi(dvb, DEBINOSWAP, BUFF1_BASE, 0, 2 ) ) {
                ddelay(1); 
                if ((jiffies - start) > ARM_WAIT_OSD) {
                        printk("outtext error\n");
                        up(&dvb->dcomlock);
                        arm_error(dvb);
                        return -1;
                }
        }
        up(&dvb->dcomlock);
        return 0;
}

static int WriteText(struct dvb_struct *dvb, u8 win, u16 x, u16 y, u8* buf)
{
        int i, ret;
        u32 start;
        int length=strlen(buf)+1;
        u16 cbuf[5] = { (COMTYPE_OSD<<8) + DText, 3, win, x, y };
        
        down_interruptible(&dvb->dcomlock);
        start = jiffies;
        while ( rdebi(dvb, DEBINOSWAP, BUFF1_BASE, 0, 2 ) ) {
                ddelay(1);
                if ((jiffies - start) > ARM_WAIT_OSD) {
                        printk("outtext error\n");
                        up(&dvb->dcomlock);
                        arm_error(dvb);
                        return -1;
                }
        }
#ifndef _NOHANDSHAKE
        start = jiffies;
        while ( rdebi(dvb, DEBINOSWAP, HANDSHAKE_REG, 0, 2 ) ) {
                ddelay(1);
                if ((jiffies - start) > ARM_WAIT_SHAKE) {
                        printk("outtext error\n");
                        up(&dvb->dcomlock);
                        arm_error(dvb);
                        return -1;
                }
        }
#endif
        for (i=0; i<length/2; i++)
                wdebi(dvb, DEBINOSWAP, BUFF1_BASE + i*2, 
                      swab16(*(u16 *)(buf+2*i)), 2);
        if (length&1)
                wdebi(dvb, DEBINOSWAP, BUFF1_BASE + i*2, 0, 2);
        ret=OutCommand(dvb, cbuf, 5);
        up(&dvb->dcomlock);
        return ret;
}

inline static int DrawLine(struct dvb_struct *dvb, u8 windownr, 
                    u16 x, u16 y, u16 dx, u16 dy, u16 color)
{
        return outcom(dvb, COMTYPE_OSD, DLine, 6,
                      windownr, x, y, dx, dy, color);
} 

inline static int DrawBlock(struct dvb_struct *dvb, u8 windownr, 
                    u16 x, u16 y, u16 dx, u16 dy, u16 color)
{
        return outcom(dvb, COMTYPE_OSD, DBox, 6,
                      windownr, x, y, dx, dy, color);
} 

inline static int HideWindow(struct dvb_struct *dvb, u8 windownr)
{
        return outcom(dvb, COMTYPE_OSD, WHide, 1, windownr);
} 

inline static int MoveWindowRel(struct dvb_struct *dvb, u8 windownr, u16 x, u16 y)
{
        return outcom(dvb, COMTYPE_OSD, WMoveD, 3, windownr, x, y);
} 

inline static int MoveWindowAbs(struct dvb_struct *dvb, u8 windownr, u16 x, u16 y)
{
        return outcom(dvb, COMTYPE_OSD, WMoveA, 3, windownr, x, y);
} 

inline static int DestroyOSDWindow(struct dvb_struct *dvb, u8 windownr)
{
        return outcom(dvb, COMTYPE_OSD, WDestroy, 1, windownr);
} 

#if 0
static void DestroyOSDWindows(struct dvb_struct *dvb)
{
        int i;

        for (i=1; i<7; i++)
                outcom(dvb, COMTYPE_OSD, WDestroy, 1, i);
} 
#endif

static inline int 
CreateOSDWindow(struct dvb_struct *dvb, u8 windownr,
                           DISPTYPE disptype, u16 width, u16 height)
{
        return outcom(dvb, COMTYPE_OSD, WCreate, 4,
                      windownr, disptype, width, height);
} 


static OSDPALTYPE bpp2pal[8]={Pal1Bit, Pal2Bit, 0, Pal4Bit, 0, 0, 0, Pal8Bit}; 
static DISPTYPE   bpp2bit[8]={BITMAP1, BITMAP2, 0, BITMAP4, 0, 0, 0, BITMAP8}; 

static inline int 
LoadBitmap(struct dvb_struct *dvb, u16 format, u16 dx, u16 dy, int inc, u8* data)
{
        int bpp;
        int i;
        int d, delta; 
        u8 c;
        DECLARE_WAITQUEUE(wait, current);
        
        if (dvb->bmp_state==BMP_LOADING) {
                add_wait_queue(&dvb->bmpq, &wait);
                while (1) {
                        set_current_state(TASK_INTERRUPTIBLE);
                        if (dvb->bmp_state!=BMP_LOADING
                            || signal_pending(current))
                                break;
                        schedule();
                }
                current->state=TASK_RUNNING;
                remove_wait_queue(&dvb->bmpq, &wait);
        }
        if (dvb->bmp_state==BMP_LOADING)
                return -1;
        dvb->bmp_state=BMP_LOADING;
        if      (format==BITMAP8) { bpp=8; delta = 1; } 
        else if (format==BITMAP4) { bpp=4; delta = 2; }
        else if (format==BITMAP2) { bpp=2; delta = 4; }
        else if (format==BITMAP1) { bpp=1; delta = 8; }
        else {
                dvb->bmp_state=BMP_NONE;
                return -1;
        }
        dvb->bmplen= ((dx*dy*bpp+7)&~7)/8; 
        dvb->bmpp=0;
        if (dvb->bmplen>32768) {
                dvb->bmp_state=BMP_NONE;
                return -1;
        }
        for (i=0; i<dy; i++) {
                if (copy_from_user(dvb->bmpbuf+1024+i*dx, data+i*inc, dx)) { 
                        dvb->bmp_state=BMP_NONE;
                        return -1;
                }
        }
        if (format != BITMAP8) {
                for (i=0; i<dx*dy/delta; i++) {
                        c = ((u8 *)dvb->bmpbuf)[1024+i*delta+delta-1];
                        for (d=delta-2; d>=0; d--) {
                                c |= (((u8 *)dvb->bmpbuf)[1024+i*delta+d] 
                                      << ((delta-d-1)*bpp));
                                ((u8 *)dvb->bmpbuf)[1024+i] = c;
                        }
                }
        }
        dvb->bmplen+=1024;
        return outcom(dvb, COMTYPE_OSD, LoadBmp, 3, format, dx, dy);
} 

static int 
BlitBitmap(struct dvb_struct *dvb, u16 win, u16 x, u16 y, u16 trans)
{
        DECLARE_WAITQUEUE(wait, current);
        
        if (dvb->bmp_state==BMP_NONE)
                return -1;
        if (dvb->bmp_state==BMP_LOADING) {
                add_wait_queue(&dvb->bmpq, &wait);
                while (1) {
                        set_current_state(TASK_INTERRUPTIBLE);
                        if (dvb->bmp_state!=BMP_LOADING
                            || signal_pending(current))
                                break;
                        schedule();
                }
                current->state=TASK_RUNNING;
                remove_wait_queue(&dvb->bmpq, &wait);
        }
        if (dvb->bmp_state==BMP_LOADED)
                return outcom(dvb, COMTYPE_OSD, BlitBmp, 4, win, x, y, trans);
        return -1;
} 

static inline int 
ReleaseBitmap(struct dvb_struct *dvb)
{
        if (dvb->bmp_state!=BMP_LOADED)
                return -1;
        dvb->bmp_state=BMP_NONE;
        return outcom(dvb, COMTYPE_OSD, ReleaseBmp, 0);
} 

static u32 RGB2YUV(u16 R, u16 G, u16 B)
{
        u16 y, u, v;
        u16 Y, Cr, Cb;

        y = R * 77 + G * 150 + B * 29;  // Luma=0.299R+0.587G+0.114B 0..65535
        u = 2048+B * 8 -(y>>5);    // Cr 0..4095
        v = 2048+R * 8 -(y>>5);    // Cb 0..4095

        Y=y/256;
        Cb=u/16;
        Cr=v/16;

        return Cr|(Cb<<16)|(Y<<8);
}

static void
OSDSetColor(struct dvb_struct *dvb, u8 color, u8 r, u8 g, u8 b, u8 blend)
{
        u16 ch, cl;
        u32 yuv;

        yuv=blend ? RGB2YUV(r,g,b) : 0;
        cl=(yuv&0xffff);
        ch=((yuv>>16)&0xffff);
        SetColor_(dvb, dvb->osdwin, bpp2pal[dvb->osdbpp[dvb->osdwin]],
                  color, ch, cl);
        SetBlend_(dvb, dvb->osdwin, bpp2pal[dvb->osdbpp[dvb->osdwin]],
                  color, ((blend>>4)&0x0f));
}

static int
OSDSetBlock(struct dvb_struct *dvb, int x0, int y0, int x1, int y1, int inc, u8 *data)
{
        uint w, h, bpp, bpl, size, lpb, bnum, brest;
        int i;

        w=x1-x0+1; h=y1-y0+1;
        if (inc<=0)
                inc=w; 
        if (w<=0 || w>720 || h<=0 || h>576) 
                return -1;
        bpp=dvb->osdbpp[dvb->osdwin]+1; 
        bpl=((w*bpp+7)&~7)/8; 
        size=h*bpl;
        lpb=(32*1024)/bpl; 
        bnum=size/(lpb*bpl);
        brest=size-bnum*lpb*bpl;

        for (i=0; i<bnum; i++) {
                LoadBitmap(dvb, bpp2bit[dvb->osdbpp[dvb->osdwin]], w, lpb, inc, data); 
                BlitBitmap(dvb, dvb->osdwin, x0, y0+i*lpb, 0);
                data+=lpb*inc; 
        }
        if (brest) {
                LoadBitmap(dvb, bpp2bit[dvb->osdbpp[dvb->osdwin]], w, brest/bpl, inc, data); 
                BlitBitmap(dvb, dvb->osdwin, x0, y0+bnum*lpb, 0);
        }
        ReleaseBitmap(dvb);
        return 0;
}

static int 
OSD_DrawCommand(struct dvb_struct *dvb, osd_cmd_t *dc)
{
        switch (dc->cmd) {
        case OSD_Close:
                DestroyOSDWindow(dvb, dvb->osdwin);
                return 0;
        case OSD_Open:
                dvb->osdbpp[dvb->osdwin]=(dc->color-1)&7;
                CreateOSDWindow(dvb, dvb->osdwin, bpp2bit[dvb->osdbpp[dvb->osdwin]],
                                dc->x1-dc->x0+1, dc->y1-dc->y0+1);
                if (!dc->data) {
                        MoveWindowAbs(dvb, dvb->osdwin, dc->x0, dc->y0);
                        SetColorBlend(dvb, dvb->osdwin);
                }
                return 0;
        case OSD_Show:
                MoveWindowRel(dvb, dvb->osdwin, 0, 0);
                return 0;
        case OSD_Hide:
                HideWindow(dvb, dvb->osdwin);
                return 0;
        case OSD_Clear:
                DrawBlock(dvb, dvb->osdwin, 0, 0, 720, 576, 0);
                return 0;
        case OSD_Fill:
                DrawBlock(dvb, dvb->osdwin, 0, 0, 720, 576, dc->color);
                return 0;
        case OSD_SetColor:
                OSDSetColor(dvb, dc->color, dc->x0, dc->y0, dc->x1, dc->y1); 
                return 0;
        case OSD_SetPalette:
        {      
                int i, len=dc->x0-dc->color+1;
                u8 colors[len*4];

                if (copy_from_user(colors, dc->data, sizeof(colors)))
                        return -EFAULT;
                for (i=0; i<len; i++)
                        OSDSetColor(dvb, dc->color+i,
                                    colors[i*4]  , colors[i*4+1],
                                    colors[i*4+2], colors[i*4+3]);
                return 0;
        }
        case OSD_SetTrans: 
                return 0;
        case OSD_SetPixel:
                DrawLine(dvb, dvb->osdwin,
                         dc->x0, dc->y0, 0, 0,
                         dc->color);
                return 0;
        case OSD_GetPixel: 
                return 0;

        case OSD_SetRow:   
                dc->y1=dc->y0;
        case OSD_SetBlock:
                OSDSetBlock(dvb, dc->x0, dc->y0, dc->x1, dc->y1, dc->color, dc->data);
                return 0;

        case OSD_FillRow:
                DrawBlock(dvb, dvb->osdwin, dc->x0, dc->y0,
                          dc->x1-dc->x0+1, dc->y1,
                          dc->color);
                return 0;
        case OSD_FillBlock:
                DrawBlock(dvb, dvb->osdwin, dc->x0, dc->y0,
                          dc->x1-dc->x0+1, dc->y1-dc->y0+1,
                          dc->color);
                return 0;
        case OSD_Line:
                DrawLine(dvb, dvb->osdwin,
                         dc->x0, dc->y0, dc->x1-dc->x0, dc->y1-dc->y0,
                         dc->color);
                return 0;
        case OSD_Query: 
                return 0;
        case OSD_Test:
                return 0;
        case OSD_Text:
        {
                char textbuf[240];
                
                if (strncpy_from_user(textbuf, dc->data, 240)<0)
                        return -EFAULT;
                textbuf[239]=0;
                if (dc->x1>3) 
                        dc->x1=3;
                SetFont(dvb, dvb->osdwin, dc->x1,
                        (u16) (dc->color&0xffff), (u16) (dc->color>>16));
                FlushText(dvb);
                WriteText(dvb, dvb->osdwin, dc->x0, dc->y0, textbuf);
                return 0;
        }
        case OSD_SetWindow:
                if (dc->x0<1 || dc->x0>7)
                        return -EINVAL;
                dvb->osdwin=dc->x0;
                return 0;
        case OSD_MoveWindow:
                MoveWindowAbs(dvb, dvb->osdwin, dc->x0, dc->y0);
                SetColorBlend(dvb, dvb->osdwin);
                return 0;
        default:
                return -EINVAL;
        }
}
#endif /* USE_OSD */


/* get version of the firmware ROM, RTSL, video ucode and ARM application  */

static void 
firmversion(struct dvb_struct *dvb)
{
        u16 buf[20];

        u16 tag = ((COMTYPE_REQUEST << 8) + ReqVersion);
        
        RequestParameter(dvb, tag, buf, 16);
        
        dvb->arm_fw=(buf[0] << 16) + buf[1];
        dvb->arm_rtsl=(buf[2] << 16) + buf[3];
        dvb->arm_vid=(buf[4] << 16) + buf[5];
        dvb->arm_app=(buf[6] << 16) + buf[7];
        dvb->avtype=(buf[8] << 16) + buf[9];

        printk("dvb%d: AV711%d - firm %08x, rtsl %08x, vid %08x, app %08x\n", 
               dvb->num, dvb->avtype, dvb->arm_fw, 
               dvb->arm_rtsl, dvb->arm_vid, dvb->arm_app);
        return;
}

static int 
waitdebi(struct dvb_struct *dvb, int adr, int state)
{
        int k;
        
        for (k=0; k<100; k++, udelay(500)) {
                if (irdebi(dvb, DEBINOSWAP, adr, 0, 2) == state) 
                        return 0;
        }
        return -1;
}

#define BOOT_STATE	(DPRAM_BASE + 0x3F8)
#define BOOT_SIZE	(DPRAM_BASE + 0x3FA)
#define BOOT_BASE	(DPRAM_BASE + 0x3FC)
#define BOOT_BLOCK	(DPRAM_BASE + 0x400)
#define BOOT_MAX_SIZE	0xc00

static int 
load_dram(struct dvb_struct *dvb, u32 *data, int len)
{
        int i;
        int blocks, rest;
        u32 base, bootblock=BOOT_BLOCK;
        
        blocks=len/BOOT_MAX_SIZE;
        rest=len % BOOT_MAX_SIZE;
        base=DRAM_START_CODE;
        
        for (i=0; i<blocks; i++) {
                if (waitdebi(dvb, BOOT_STATE, BOOTSTATE_BUFFER_EMPTY) < 0)
                        return -1;
                dprintk("Writing DRAM block %d\n",i);
                iwdebi(dvb, DEBISWAB, bootblock,
                       i*(BOOT_MAX_SIZE)+(u32)data,
                       BOOT_MAX_SIZE);
                bootblock^=0x1400;
                iwdebi(dvb, DEBISWAB, BOOT_BASE, cpu_to_be32(base), 4);
                iwdebi(dvb, DEBINOSWAP, BOOT_SIZE, BOOT_MAX_SIZE, 2);
                iwdebi(dvb, DEBINOSWAP, BOOT_STATE, BOOTSTATE_BUFFER_FULL, 2);
                base+=BOOT_MAX_SIZE;
        }
        
        if (rest > 0) {
                if (waitdebi(dvb, BOOT_STATE, BOOTSTATE_BUFFER_EMPTY) < 0)
                        return -1;
                if (rest>4)
                        iwdebi(dvb, DEBISWAB, bootblock, i*(BOOT_MAX_SIZE)+(u32)data, rest);
                else
                        iwdebi(dvb, DEBISWAB, bootblock, i*(BOOT_MAX_SIZE)-4+(u32)data, rest+4);
                
                iwdebi(dvb, DEBISWAB, BOOT_BASE, cpu_to_be32(base), 4);
                iwdebi(dvb, DEBINOSWAP, BOOT_SIZE, rest, 2);
                iwdebi(dvb, DEBINOSWAP, BOOT_STATE, BOOTSTATE_BUFFER_FULL, 2);
        }
        if (waitdebi(dvb, BOOT_STATE, BOOTSTATE_BUFFER_EMPTY) < 0)
                return -1;
        iwdebi(dvb, DEBINOSWAP, BOOT_SIZE, 0, 2);
        iwdebi(dvb, DEBINOSWAP, BOOT_STATE, BOOTSTATE_BUFFER_FULL, 2);
        if (waitdebi(dvb, BOOT_STATE, BOOTSTATE_BOOT_COMPLETE) < 0)
                return -1;
        return 0;
}


#define INC_BOOT
//#define INC_FIRM
//#define FIRM_MOD

static u8 
bootcode[] __initdata = {
        0xea, 0x00, 0x00, 0x0e, 0xe1, 0xb0, 0xf0, 0x0e, /* 0x0000 */
        0xe2, 0x5e, 0xf0, 0x04, 0xe2, 0x5e, 0xf0, 0x04,
        0xe2, 0x5e, 0xf0, 0x08, 0xe2, 0x5e, 0xf0, 0x04,
        0xe2, 0x5e, 0xf0, 0x04, 0xe2, 0x5e, 0xf0, 0x04,
        0x2c, 0x00, 0x00, 0x24, 0x00, 0x00, 0x00, 0x0c,
        0x00, 0x00, 0x00, 0x00, 0x2c, 0x00, 0x00, 0x34,
        0x00, 0x00, 0x00, 0x00, 0xa5, 0xa5, 0x5a, 0x5a,
        0x00, 0x1f, 0x15, 0x55, 0x00, 0x00, 0x00, 0x09,
        0xe5, 0x9f, 0xd0, 0x5c, 0xe5, 0x9f, 0x40, 0x54, /* 0x0040 */
        0xe3, 0xa0, 0x00, 0x00, 0xe5, 0x84, 0x00, 0x00,
        0xe5, 0x84, 0x00, 0x04, 0xe1, 0xd4, 0x10, 0xb0,
        0xe3, 0x51, 0x00, 0x00, 0x0a, 0xff, 0xff, 0xfc,
        0xe1, 0xa0, 0x10, 0x0d, 0xe5, 0x94, 0x30, 0x04,
        0xe1, 0xd4, 0x20, 0xb2, 0xe2, 0x82, 0x20, 0x3f,
        0xe1, 0xb0, 0x23, 0x22, 0x03, 0xa0, 0x00, 0x02,
        0xe1, 0xc4, 0x00, 0xb0, 0x0a, 0xff, 0xff, 0xf4,
        0xe8, 0xb1, 0x1f, 0xe0, 0xe8, 0xa3, 0x1f, 0xe0, /* 0x0080 */
        0xe8, 0xb1, 0x1f, 0xe0, 0xe8, 0xa3, 0x1f, 0xe0,
        0xe2, 0x52, 0x20, 0x01, 0x1a, 0xff, 0xff, 0xf9,
        0xe2, 0x2d, 0xdb, 0x05, 0xea, 0xff, 0xff, 0xec,
        0x2c, 0x00, 0x03, 0xf8, 0x2c, 0x00, 0x04, 0x00,
};

#ifdef INC_FIRM
#include "dvb_firm.h"
#endif

#ifdef FIRM_MOD
unsigned char *dvb_dpram_addr, *dvb_root_addr;
int dvb_dpram_len, dvb_root_len;
static int firm_mod=-1;
#endif

__init static int 
bootarm(struct dvb_struct *dvb)
{
        u32 *lfirm;
        int len;
        int result;
        u32 tmp;

        setgpio(dvb, RESET_LINE, GPIO_OUTLO);

        /* disable DEBI and GPIO irq */
        saa7146_write(dvb->saa->mem, IER, 
                      saa7146_read(dvb->saa->mem, IER) & 
                      ~(MASK_19 | MASK_03));
        saa7146_write(dvb->saa->mem, ISR, (MASK_19 | MASK_03));

        /* enable DEBI */
        saa7146_write(dvb->saa->mem, MC1, 0x08800880);
        saa7146_write(dvb->saa->mem, DD1_STREAM_B, 0x00000000);
        saa7146_write(dvb->saa->mem, MC2, (MASK_09 | MASK_25 | MASK_10 | MASK_26));
        
        /* test DEBI */
        iwdebi(dvb, DEBISWAP, DPRAM_BASE, 0x76543210, 4);
        if ((tmp=irdebi(dvb, DEBINOSWAP, DPRAM_BASE, 0, 4))!=0x10325476) {
                printk("dvb: debi test in bootarm() failed: "
                       "%08x != %08x\n", tmp, 0x10325476);;
                return -1;
        }
        dprintk("bootarm: debi test OK\n");;

        /* boot */
        dprintk("bootarm: load boot code\n");
#ifndef INC_BOOT
        if (
#ifdef __DVB_PACK__
            !(len=firmread("Boot_up.axf", (char**) &lfirm)) &&
#endif
            !(len=firmread(DVB_FIRM_PATH "Boot_up.axf", (char**) &lfirm)))
                        return -1;
#endif

        setgpio(dvb, ARM_IRQ_LINE, GPIO_IRQLO);
        //setgpio(dvb, DEBI_DONE_LINE, GPIO_INPUT);
        //setgpio(dvb, 3, GPIO_INPUT);

#ifdef INC_BOOT  
        iwdebi(dvb, DEBISWAB, DPRAM_BASE, (u32) bootcode, sizeof(bootcode));
#else
        iwdebi(dvb, DEBISWAB, DPRAM_BASE, (u32) lfirm, len);
#endif
        iwdebi(dvb, DEBINOSWAP, BOOT_STATE, BOOTSTATE_BUFFER_FULL, 2);
        
#ifndef INC_BOOT  
        vfree(lfirm);
#endif
        wait_for_debi_done(dvb);
        setgpio(dvb, RESET_LINE, GPIO_OUTHI);
        current->state=TASK_INTERRUPTIBLE;
        schedule_timeout(HZ);
        
        dprintk("bootarm: load dram code\n");
#ifdef INC_FIRM
        if (readfirm) {
#else
#ifdef FIRM_MOD
        if (firm_mod<0 || !dvb_root_len) {
#endif
#endif
                if (
#ifdef __DVB_PACK__
                        !(len=firmread("arm_firm/Root", (char**) &lfirm)) &&
                        !(len=firmread("Root", (char**) &lfirm)) &&
#endif
                        !(len=firmread(DVB_FIRM_PATH "Root", (char**) &lfirm)))
                        return -1;
                result=load_dram(dvb, lfirm, len);
                vfree(lfirm);
                if (result<0)
                        return -1;
#ifdef INC_FIRM
        } else
                if (load_dram(dvb, (u32 *)Root, sizeof(Root))<0)
                        return -1;
#else
#ifdef FIRM_MOD
        } else
                if (load_dram(dvb, (u32 *)dvb_root_addr, dvb_root_len)<0)
                        return -1;
#endif
#endif
        setgpio(dvb, RESET_LINE, GPIO_OUTLO);
        mdelay(1);
        

        dprintk("bootarm: load dpram code\n");
#ifdef INC_FIRM
        if (readfirm) {
#else
#ifdef FIRM_MOD
        if (firm_mod<0 || !dvb_dpram_len) {
                printk("no dpram_len\n");
#endif
#endif
                if (
#ifdef __DVB_PACK__
                        !(len=firmread("arm_firm/Dpram", (char**) &lfirm)) &&
                        !(len=firmread("Dpram", (char**) &lfirm)) &&
#endif
                        !(len=firmread(DVB_FIRM_PATH "Dpram", (char**) &lfirm)))
                        return -1;
                iwdebi(dvb, DEBISWAB, DPRAM_BASE, (u32) lfirm, len);
                vfree(lfirm);
#ifdef INC_FIRM
        } else 
                iwdebi(dvb, DEBISWAB, DPRAM_BASE, (u32) Dpram, sizeof(Dpram));
#else
#ifdef FIRM_MOD
        } else
                iwdebi(dvb, DEBISWAB, DPRAM_BASE, (u32) dvb_dpram_addr, 
                       dvb_dpram_len);
#endif
#endif        
        wait_for_debi_done(dvb);

        setgpio(dvb, RESET_LINE, GPIO_OUTHI);
        mdelay(800);

        //ARM_ClearIrq(dvb); 
        ARM_ResetMailBox(dvb); 
        saa7146_write(dvb->saa->mem, ISR, (MASK_19 | MASK_03));
        saa7146_write(dvb->saa->mem, IER, 
                      saa7146_read(dvb->saa->mem, IER) | MASK_03 );
        firmversion(dvb);
        return 0;
}

static int
SetPIDs(struct dvb_struct *dvb, u16 vpid, u16 apid, u16 ttpid, 
        u16 subpid, u16 pcrpid)
{
        return outcom(dvb, COMTYPE_PIDFILTER, MultiPID, 5, 
                      pcrpid, vpid, apid, ttpid, subpid);
}

static void
ChangePIDs(struct dvb_struct *dvb, u16 vpid, u16 apid, u16 ttpid, 
        u16 subpid, u16 pcrpid)
{

        if (!(vpid&0x8000))  dvb->front.video_pid=vpid;
        if (!(apid&0x8000))  dvb->front.audio_pid=apid;
        if (!(ttpid&0x8000)) dvb->front.tt_pid=ttpid;
        if (!(pcrpid&0x8000)) dvb->front.pcr_pid=pcrpid;

        if (dvb->mon_tuning!=3) return;
        outcom(dvb, COMTYPE_PIDFILTER, MultiPID, 5, 
               pcrpid, vpid, apid, ttpid, subpid);
}

static void
SetFEPIDs(struct dvb_struct *dvb)
{
        SetPIDs(dvb, dvb->front.video_pid, 
                dvb->front.audio_pid, dvb->front.tt_pid, 0, 
                dvb->front.pcr_pid);
}

static void 
SetMode(struct dvb_struct *dvb, int mode)
{
        outcom(dvb, COMTYPE_ENCODER, LoadVidCode, 1, mode);
        
        if (!dvb->playing) {
                ChangePIDs(dvb, dvb->front.video_pid, 
                        dvb->front.audio_pid, dvb->front.tt_pid, 
                           0, dvb->front.pcr_pid);
                outcom(dvb, COMTYPE_PIDFILTER, Scan, 0);
        }
}

inline static void 
TestMode(struct dvb_struct *dvb, int mode)
{
        outcom(dvb, COMTYPE_ENCODER, SetTestMode, 1, mode);
}

inline static void 
VidMode(struct dvb_struct *dvb, int mode)
{
        outcom(dvb, COMTYPE_ENCODER, SetVidMode, 1, mode);
}
           

static int inline
vidcom(struct dvb_struct *dvb, u32 com, u32 arg)
{
        return outcom(dvb, 0x80, 0x02, 4, 
                      (com>>16), (com&0xffff), 
                      (arg>>16), (arg&0xffff));
}

static int inline
audcom(struct dvb_struct *dvb, u32 com)
{
        return outcom(dvb, 0x80, 0x03, 4, 
                      (com>>16), (com&0xffff));
}

inline static void 
Set22K(struct dvb_struct *dvb, int state)
{
        if (dvb->card_type==DVB_CARD_TT_SIEMENS)
                outcom(dvb, COMTYPE_AUDIODAC, (state ? ON22K : OFF22K), 0);
        if (dvb->card_type==DVB_CARD_TT_BUDGET)
                setgpio(dvb, 3, (state ? GPIO_OUTHI : GPIO_OUTLO));
}


/* Diseqc functions only for TT Budget card */
/* taken from the Skyvision DVB driver by 
   Ralph Metzler <rjkm@metzlerbros.de> */


inline static void 
DiseqcSendBit(struct dvb_struct *dvb, int data)
{
        setgpio(dvb, 3, GPIO_OUTHI);
        udelay(data ? 500 : 1000);
        setgpio(dvb, 3, GPIO_OUTLO);
        udelay(data ? 1000 : 500);
}

static void 
DiseqcSendByte(struct dvb_struct *dvb, int data)
{
        int i, par=1, d;

        for (i=7; i>=0; i--) 
        {
                d=(data>>i)&1;
                par^=d;
                DiseqcSendBit(dvb, d);
        }
        DiseqcSendBit(dvb, par);
}

static void 
DiseqcSendMsg(struct dvb_struct *dvb, int frame, int adr, int com, int data, int swi)
{
        setgpio(dvb, 3, GPIO_OUTLO);

        mdelay(16);

        DiseqcSendByte(dvb, frame);
        DiseqcSendByte(dvb, adr);
        DiseqcSendByte(dvb, com);
        DiseqcSendByte(dvb, data);

        mdelay(16);
        
        if (swi) 
                DiseqcSendByte(dvb, 0xff);
        else {
                setgpio(dvb, 3, GPIO_OUTHI);
                udelay(12500);
                setgpio(dvb, 3, GPIO_OUTLO);
        }

        ddelay(2);

        Set22K(dvb, dvb->front.ttk);
}


inline static void
SetDiSEqC(struct dvb_struct *dvb, int input)
{
        dprintk("setdis: %d\n", input);
        if (input<0 || input>15)
                return;
        if (dvb->card_type==DVB_CARD_TT_SIEMENS) {
                outcom(dvb, COMTYPE_AUDIODAC, SendDiSEqC, 6, 4,
                       (input&4) ? 1 : 0,
                       0xe0, 0x10, 0x38, 0xf0+input);
                outcom(dvb, COMTYPE_AUDIODAC, SendDiSEqC, 6, 4,
                       (input&4) ? 1 : 0,
                       0xe1, 0x10, 0x38, 0xf0+input);
        } else {
                DiseqcSendMsg(dvb, 0xe0, 0x10, 0x38, 0xf0+input, (input&4));
                DiseqcSendMsg(dvb, 0xe1, 0x10, 0x38, 0xf0+input, (input&4));
        }
}

inline static int
SendDiSEqCMsg(struct dvb_struct *dvb, int len, u8 *msg, int burst)
{
        int i;

        switch (dvb->card_type) {
        case DVB_CARD_TT_SIEMENS:
        {
                u16 buf[18] = { ((COMTYPE_AUDIODAC << 8) + SendDiSEqC),
                                16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
                
                if (len>10)
                        len=10;
                buf[1] = len+2;
                buf[2] = len;
                
                if (burst!=-1)
                  buf[3]=burst ? 0xff : 0;
                else
                  buf[3]=0xffff;
                
                for (i=0; i<len; i++)
                  buf[i+4]=msg[i];
                
                SOutCommand(dvb, buf, 18);
                break;
        }

        case DVB_CARD_TT_BUDGET:
                setgpio(dvb, 3, GPIO_OUTLO);
                
                mdelay(16);
                
                for (i=0; i<len; i++)
                        DiseqcSendByte(dvb, msg[i]);

                mdelay(16);
                
                if (burst!=-1) {
                        if (burst) 
                                DiseqcSendByte(dvb, 0xff);
                        else {
                                setgpio(dvb, 3, GPIO_OUTHI);
                                udelay(12500);
                                setgpio(dvb, 3, GPIO_OUTLO);
                        }
                        
                        ddelay(2);
                }
                Set22K(dvb, dvb->front.ttk);

                break;

        default:
                return -1;
        }
        return 0;
}

/****************************************************************************
 * I2C client commands
 ****************************************************************************/


static int demod_command(struct dvb_struct *dvb, unsigned int cmd, void *arg)
{
        if (!dvb->demod)
                return -1;
        return dvb->demod->driver->command(dvb->demod, cmd, arg);
}

static inline u8
tuner_stat(struct dvb_struct *dvb, u8 id)
{
        struct i2c_adapter *adap=dvb->i2cbus;
        u8 mm1[] = {0x00};
        struct i2c_msg msgs[1];

        msgs[0].flags=I2C_M_RD;
        msgs[0].addr=id/2;
        msgs[0].len=1;
        msgs[0].buf=mm1;
        if (i2c_transfer(adap, msgs, 1)<0)
          printk("dvb: tuner_stat error\n");

        return mm1[0];
}

static int tuner_command(struct dvb_struct *dvb, unsigned int cmd, void *arg)
{
        if (!dvb->tuner)
                return -1;
        if (cmd == TUNER_SET_TVFREQ &&
            dvb->demod && 
            dvb->demod_type==DVB_DEMOD_STV0299) {
                u8 msg[2]={ 0x05, 0x85 }; 
                
                demod_command(dvb, DVB_WRITEREG, msg); 
                dvb->tuner->driver->command(dvb->tuner, cmd, arg);
                
                //demod_command(dvb, DVB_WRITEREG, msg); 
                //printk("tunerstat: %02x  freq=%08x\n", tuner_stat(dvb, 0xc2), *(unsigned int *)arg);
                
                return 0;
        }

        return dvb->tuner->driver->command( dvb->tuner, cmd, arg);
}

static void
GetFront(struct dvb_struct *dvb)
{
        demod_command(dvb, DVB_GET_FRONTEND, &dvb->front); 
}

/****************************************************************************
 * Tuning
 ****************************************************************************/

static const uint8_t fectab[8]={8,0,1,2,4,6,0,8};
static const uint8_t fectab2[9]={1,2,3,-1,4,-1,5,-1,0};


static int
fe_lock(struct dvb_struct *dvb)
{
        int lock;

        demod_command(dvb, DVB_GET_FRONTEND, &dvb->front); 
        switch (dvb->dvbtype) {
        case DVB_S:
                lock=(dvb->front.sync&0x10 ? 1 : 0);
                break;
        case DVB_C:
                lock=(dvb->front.sync&0x08 ? 1 : 0);
                break;
        default:
                lock=0;
                break;
        }
        dvb->mon_lock=lock;
        return lock;
}

static void 
mon_timer(unsigned long data)
{
	struct dvb_struct *dvb = (struct dvb_struct*) data;

	wake_up_interruptible(&dvb->mon_wait);
}

static unsigned long 
mon_afc(struct dvb_struct *dvb)
{
        s32 afc;

        GetFront(dvb);
        afc=dvb->front.afc;

        dprintk("dvb_mon(%d): mon_afc freq:%d curfreq:%d "
                "afc:%d sync:%x nest:%x vber:%d\n", 
                dvb->num,
                dvb->front.freq, dvb->front.curfreq, 
                dvb->front.afc, dvb->front.sync, 
                dvb->front.nest, dvb->front.vber);

        if (!fe_lock(dvb)) {
                dvb->mon_tuning=1; // re-tune
                dvb->mon_delay=HZ/2;
                return -1;
        }
        dvb->mon_delay=HZ*60;
        if (dvb->dvbtype==DVB_C)
                return 0;
#if 0
        /* this does not work well for me, after the first adjustment I
           lose sync in most cases */

        if (!afc) {
                dvb->mon_delay=HZ*60;
                return 0;  
        }
        dvb->mon_delay=HZ*10;
        
        dvb->front.curfreq -= (afc/2); 
        tuner_command(dvb, TUNER_SET_TVFREQ, &dvb->front.curfreq); 

        /* urghh, this prevents loss of lock but is visible as artifacts */
        demod_command(dvb, DVB_RESET, 0); 
#endif
	return 0;
}

static int 
mon_complete(struct dvb_struct *dvb)
{
        switch (dvb->dvbtype) {
        case DVB_S:
        {
                struct qpskEvent ev;
                
                ev.type=FE_COMPLETION_EV;
                ev.u.completionEvent.iFrequency=dvb->front.freq/1000;
                ev.u.completionEvent.SymbolRate=dvb->front.srate;
                ev.u.completionEvent.FEC_inner=fectab2[dvb->front.fec];
                add_qpsk_event(&dvb->qpsk, &ev);
                break;
        }
        case DVB_C:
        {
                struct qamEvent ev;
                
                ev.type=FE_COMPLETION_EV;
                ev.u.completionEvent.Frequency=dvb->front.freq;
                ev.u.completionEvent.SymbolRate=dvb->front.srate;
                ev.u.completionEvent.FEC_inner=fectab2[dvb->front.fec];
                ev.u.completionEvent.QAM=dvb->front.qam;
                add_qam_event(&dvb->qam, &ev);
                break;
        }
        default:
                break;
        }
        dvb->mon_tuning=3;
        dvb->mon_delay=HZ*10;
        SetFEPIDs(dvb);
        outcom(dvb, COMTYPE_PIDFILTER, Scan, 0);
        return 0;
}

static int
mon_fail(struct dvb_struct *dvb)
{
        switch (dvb->dvbtype) {
        case DVB_S:
        {
                struct qpskEvent ev;
                
                dvb->mon_tuning=0;
                ev.type=FE_FAILURE_EV;
                add_qpsk_event(&dvb->qpsk, &ev);
                break;
        }
        case DVB_C:
        {
                struct qamEvent ev;

                dvb->mon_tuning=0;
                ev.type=FE_FAILURE_EV;
                add_qam_event(&dvb->qam, &ev);
                break;
        }
        }
        return -1;
}

static int 
mon_zigzag(struct dvb_struct *dvb)
{
        int i=dvb->mon_fstate;
        u32 sfreq=dvb->front.curfreq=dvb->front.freq;
        u32 soff=dvb->front.srate/16;
        
        if (fe_lock(dvb)) 
                return mon_complete(dvb);
        if (i==20) 
                return mon_fail(dvb);
        if (i&1) 
                sfreq=dvb->front.freq+soff*(i/2);
        else
                sfreq=dvb->front.freq-soff*(i/2);
        tuner_command(dvb, TUNER_SET_TVFREQ, &sfreq); 
        dprintk("i=%d freq=%d\n", i, sfreq);
        dvb->front.curfreq=sfreq;
        i++;
        dvb->mon_fstate=i;
        if (dvb->dvbtype==DVB_C)
                dvb->mon_delay=HZ/2;
        else
                dvb->mon_delay=HZ/20;
        if (dvb->demod_type==DVB_DEMOD_STV0299) {
                dvb->mon_delay=HZ/20;
                demod_command(dvb, DVB_RESET, 0); 
        }
        return i;
}

static int 
mon_tune(struct dvb_struct *dvb)
{
        SetPIDs(dvb, 0, 0, 0, 0, 0);
        
        switch (dvb->dvbtype) {
        case DVB_S:
                if (dvb->front.srate!=dvb->new_srate || 
                    dvb->front.fec!=dvb->new_fec) {
                        dvb->front.srate=dvb->new_srate;
                        dvb->front.fec=dvb->new_fec;
                        demod_command(dvb, DVB_SET_FRONTEND, &dvb->front); 
                        //if (dvb->demod_type==DVB_DEMOD_STV0299) 
                                demod_command(dvb, DVB_RESET, 0); 
                } 
                if (dvb->front.freq==dvb->new_freq && fe_lock(dvb))
                        return mon_complete(dvb);

                dvb->front.freq=dvb->new_freq;
                tuner_command(dvb, TUNER_SET_TVFREQ, &dvb->front.freq); 
                if (dvb->demod_type==DVB_DEMOD_STV0299) 
                        demod_command(dvb, DVB_RESET, 0); 
                dprintk("freq=%d\n", dvb->new_freq);
                mdelay(10);
                if (fe_lock(dvb)) 
                        return mon_complete(dvb);
                dvb->mon_tuning=2;
                dvb->mon_fstate=1;
                dvb->mon_delay=HZ/10;
                dvb->front.curfreq=dvb->front.freq;
                return 0;
        case DVB_C:
                if (dvb->front.srate!=dvb->new_srate || 
                    dvb->front.fec!=dvb->new_fec ||
                    dvb->front.qam!=dvb->new_qam ) {
                        dvb->front.srate=dvb->new_srate;
                        dvb->front.fec=dvb->new_fec;
                        dvb->front.qam=dvb->new_qam;
                        demod_command(dvb, DVB_SET_FRONTEND, &dvb->front); 
                } 
                dvb->front.freq=dvb->new_freq;
                tuner_command(dvb, TUNER_SET_TVFREQ, &dvb->front.freq); 
                mdelay(30);
                if (fe_lock(dvb)) 
                        return mon_complete(dvb);
                dvb->mon_tuning=2;
                dvb->mon_fstate=0;
                dvb->mon_delay=HZ/2;
                dvb->front.curfreq=dvb->front.freq;
                return 0;
        default:
                return 0;
        }
}

static void
mon_do_tune(struct dvb_struct *dvb)
{
        down_interruptible(&dvb->mon_sem);
        dvb->mon_tuning=1;
        mod_timer(&dvb->mon_timer, jiffies);
        up(&dvb->mon_sem);
}

static void
mon_stop(struct dvb_struct *dvb)
{
        down_interruptible(&dvb->mon_sem);
        dvb->mon_tuning=0;
        mod_timer(&dvb->mon_timer, jiffies);
        SetPIDs(dvb, 0, 0, 0, 0, 0);
        up(&dvb->mon_sem);
}

static int mon_thread(void *data)
{
	struct dvb_struct *dvb = data;
    
#ifdef CONFIG_SMP
	lock_kernel();
#endif
	daemonize();
	sigfillset(&current->blocked);
	strcpy(current->comm,"dvb_mon");
	dvb->mon_thread = current;
#ifdef CONFIG_SMP
	unlock_kernel();
#endif
	dprintk("dvb: monitor daemon started\n");

	for (;;) {
		if (dvb->mon_rmmod)
			goto done;
                interruptible_sleep_on(&dvb->mon_wait);
                dprintk("dvb_mon: wakeup\n");
		if (dvb->mon_rmmod || signal_pending(current))
			goto done;
                down_interruptible(&dvb->mon_sem);

                switch (dvb->mon_tuning) {
                case 1:
                        mon_tune(dvb);
                        break;
                case 2:
                        mon_zigzag(dvb);
                        break;
                case 3:
                        mon_afc(dvb);
                        break;
                default:
                        dvb->mon_delay=HZ;

                        break;
                }
                if (dvb->mon_tuning)
                        mod_timer(&dvb->mon_timer, jiffies+dvb->mon_delay);
                up(&dvb->mon_sem);
	}

done:
	dprintk("dvb_mon: exit\n");
	dvb->mon_thread = NULL;
	return 0;
}




/*************************************************************/

static inline int
i2c_writereg(struct dvb_struct *dvb, u8 id, u8 reg, u8 val)
{
        u8 msg[2]={ reg, val }; 
        struct i2c_adapter *adap=dvb->i2cbus;
        struct i2c_msg msgs;

        msgs.flags=0;
        msgs.addr=id/2;
        msgs.len=2;
        msgs.buf=msg;
        return i2c_transfer(adap, &msgs, 1);
}

static inline int
msp_writereg(struct dvb_struct *dvb, u8 dev, u16 reg, u16 val)
{
        u8 msg[5]={ dev, reg>>8, reg&0xff, val>>8 , val&0xff }; 
        struct i2c_adapter *adap=dvb->i2cbus;
        struct i2c_msg msgs;

        msgs.flags=0;
        msgs.addr=0x40;
        msgs.len=5;
        msgs.buf=msg;
        return i2c_transfer(adap, &msgs, 1);
}

static inline u8
i2c_readreg(struct dvb_struct *dvb, u8 id, u8 reg)
{
        struct i2c_adapter *adap=dvb->i2cbus;
        u8 mm1[] = {0x00};
        u8 mm2[] = {0x00};
        struct i2c_msg msgs[2];

        msgs[0].flags=0;
        msgs[1].flags=I2C_M_RD;
        msgs[0].addr=msgs[1].addr=id/2;
        mm1[0]=reg;
        msgs[0].len=1; msgs[1].len=1;
        msgs[0].buf=mm1; msgs[1].buf=mm2;
        i2c_transfer(adap, msgs, 2);

        return mm2[0];
}


static inline void 
demod_writereg(struct dvb_struct *dvb, u8 adr, u8 val)
{
        u8 msg[2]={ adr, val }; 
  
        demod_command(dvb, DVB_WRITEREG, &msg); 
}

static inline u8
demod_readreg(struct dvb_struct *dvb, u8 adr)
{
        u8 msg[2]={ adr, 0 }; 
  
        demod_command(dvb, DVB_READREG, &msg); 
        return msg[1];
}


static void 
SetVoltage(struct dvb_struct *dvb, int power, int voltage)
{
        if (!dvb->demod)
                return;

        if (dvb->demod_type==DVB_DEMOD_STV0299)
                demod_writereg(dvb, 0x0c, 1|(power ? 0x40 : 0)|(voltage ? 0x10 : 0));

        if (dvb->demod_type==DVB_DEMOD_VES1893)
                demod_writereg(dvb, 0x1f, (power ? 0x20 : 0)|(voltage ? 0x10 : 0));
}

static void
InitFront(struct dvb_struct *dvb)
{
        struct qpskParameters para;
        struct qamParameters qampara;

        dprintk("initfront tunertype=%d dvbtype=%d\n", dvb->tuner_type,
               dvb->dvbtype);

        memset(&(dvb->front), 0x00, sizeof(struct frontend));

        demod_command(dvb, DVB_INIT_FRONTEND, 0); 

        dvb->powerstate=OST_POWER_ON;
        dvb->front.type=dvb->dvbtype;
        dvb->front.power=1; 
        dvb->front.pcr_pid=0;

        if (dvb->tuner_type>=0)
                tuner_command(dvb, TUNER_SET_TYPE, &dvb->tuner_type); 

        switch (dvb->dvbtype) {
        case DVB_S:
                saa7146_write(dvb->saa->mem, MC2, 
                              (MASK_09 | MASK_25 | MASK_10 | MASK_26));
                if (!init_chan)
                        break;

                if (init_chan==2) {
                        // Astra n-tv default
                        ChangePIDs(dvb, 162, 96, 0, 0, 0);
                        secSetTone(dvb, SEC_TONE_ON);
                        secSetVoltage(dvb, SEC_VOLTAGE_13);
                        para.iFrequency=12666000-10600000;
                        para.SymbolRate=22000000;
                        para.FEC_inner=0;
                        qpsk_tune(dvb, &para);
                } else {
                        ChangePIDs(dvb, 255, 256, 0, 0, 0);
                        secSetTone(dvb, SEC_TONE_ON);
                        secSetVoltage(dvb, SEC_VOLTAGE_13);
                        para.iFrequency=12480000-10600000;
                        para.SymbolRate=27500000;
                        para.FEC_inner=0;
                        qpsk_tune(dvb, &para);
                }
                SetVolume(dvb, 0xff, 0xff);
		break;

        case DVB_C:
                demod_writereg(dvb, 0x0f, 0x28); //use 0x68 for analog 

                if (!init_chan)
                        break;
                
                ChangePIDs(dvb, 0x262, 0x26c, 0, 0, 0x262);
                qampara.Frequency=394000000;
                qampara.SymbolRate=6900000;
                qampara.FEC_inner=0;
                qampara.QAM=QAM_64;
                qam_tune(dvb, &qampara);
                break;
        }
        if (dvb->card_type==DVB_CARD_TT_SIEMENS) {
                VidMode(dvb, vidmode);
                //iwdebi(dvb, DEBINOSWAP, RX_BUFF, 0, 2);
                //iwdebi(dvb, DEBINOSWAP, TX_BUFF, 0, 2);
                //ARM_ResetMailBox(dvb);
        }
}




/****************************************************************************
 * I/O buffer management and control
 ****************************************************************************/

static int sw2mode[16] = {
        VIDEO_MODE_PAL, VIDEO_MODE_PAL, VIDEO_MODE_PAL, VIDEO_MODE_PAL,
        VIDEO_MODE_NTSC, VIDEO_MODE_NTSC, VIDEO_MODE_PAL, VIDEO_MODE_NTSC,
        VIDEO_MODE_PAL, VIDEO_MODE_PAL, VIDEO_MODE_PAL, VIDEO_MODE_PAL,
        VIDEO_MODE_PAL, VIDEO_MODE_PAL, VIDEO_MODE_PAL, VIDEO_MODE_PAL,
};

static void
get_video_format(struct dvb_struct *dvb, u8 *buf, int count)
{
        int i;
	int hsize,vsize;
        int sw;
        u8 *p;

        if (dvb->sinfo)
                return;
        for (i=7; i<count-10; i++) {
                p=buf+i;
                if (p[0] || p[1] || p[2]!=0x01 || p[3]!=0xb3)
                        continue;
                p+=4;
                hsize = ((p[1] &0xF0) >> 4) | (p[0] << 4);
                vsize = ((p[1] &0x0F) << 8) | (p[2]);
                sw = (p[3]&0x0F);
                SetMode(dvb, sw2mode[sw]);
                printk("dvb: playback %dx%d fr=%d\n", hsize, vsize, sw);
                dvb->sinfo=1;
                break;
        }
}

static void
play_video_cb(u8 *buf, int count, void *priv)
{
        struct dvb_struct *dvb=(struct dvb_struct *) priv;

        if ((buf[3]&0xe0)==0xe0) {
                get_video_format(dvb, buf, count);
                dvb_buf_write(&dvb->avout, buf, count, 0, 0);
        } else
                dvb_buf_write(&dvb->aout, buf, count, 0, 0);
}

static void
play_audio_cb(u8 *buf, int count, void *priv)
{
        struct dvb_struct *dvb=(struct dvb_struct *) priv;
        
        dvb_buf_write(&dvb->aout, buf, count, 0, 0);
}

#define FREE_COND (dvb_buf_free(&dvb->avout)>=20*1024 && dvb_buf_free(&dvb->aout)>=20*1024)

static ssize_t 
dvb_play(struct dvb_struct *dvb, const u8 *buf,
         unsigned long count, int nonblock, int type, int umem)
{
        unsigned long todo = count, n;

        if (!dvb->kbuf[type])
                return -ENOBUFS;
        if (nonblock && !FREE_COND)
                return -EWOULDBLOCK;
                
        while (todo>0) {
                if (!FREE_COND) {
                        if (nonblock)
                                return count-todo;
                        if (wait_event_interruptible(dvb->avout.queue,
                                                     FREE_COND))
                        	return count-todo;
                } 
		n=todo;
                if (n>IPACKS*2)
                        n=IPACKS*2;
                if (umem) {
                        if (copy_from_user(dvb->kbuf[type], buf, n)) 
                                return -EFAULT;
                        instant_repack(dvb->kbuf[type], n, &dvb->ipack[type]);
                } else
                        instant_repack((u8 *)buf, n, &dvb->ipack[type]);
                todo -= n;
                buf += n;
        }
	return count-todo;
}

static ssize_t 
dvb_aplay(struct dvb_struct *dvb, const u8 *buf,
         unsigned long count, int nonblock, int type)
{
        unsigned long todo = count, n;

        if (!dvb->kbuf[type])
                return -ENOBUFS;
        if (nonblock && dvb_buf_free(&dvb->aout)<20*1024)
                return -EWOULDBLOCK;
                
        while (todo>0) {
                if (dvb_buf_free(&dvb->aout)<20*1024) {
                        if (nonblock)
                                return count-todo;
                        if (wait_event_interruptible(dvb->aout.queue,
                                                     (dvb_buf_free(&dvb->aout)>=
                                                      20*1024)))
                        	return count-todo;
                } 
		n=todo;
                if (n>IPACKS*2)
                        n=IPACKS*2;
                if (copy_from_user(dvb->kbuf[type], buf, n)) 
                        return -EFAULT;
                instant_repack(dvb->kbuf[type], n, &dvb->ipack[type]);
                todo -= n;
                buf += n;
        }
	return count-todo;
}

#if 0
static int find_video_format(trans_t *p, u8 *buf)
{
	u8 headr[4];
	int found=0;
        int c = 0;
	int form = -1;
	int hsize,vsize;
        int sw;
        int w = p->which_pesto;
        u32 bit_rate;

	while (c < MAX_AV-8 && !found) {
		memcpy(headr,buf+c,4);
                if (headr[0] == 0x00 && headr[1] == 0x00 && headr[2] == 0x01)
                        if ( headr[3] == 0xB3 ){ 
				found = 1;
                                c+=4;
			}
		if (! found)  c++;
	}
        
        if (!found) return -1;
        printk("MPEG%d: video: ",p->pesto[w].mpeg);


        memcpy(headr,buf+c,4);
        c += 4;

	hsize = ((headr[1] &0xF0) >> 4) | (headr[0] << 4);
	vsize = ((headr[1] &0x0F) << 8) | (headr[2]);
        p->pesto[w].hsize = hsize;
        p->pesto[w].vsize = vsize;

        sw = (int)((headr[3]&0xF0) >> 4) ;

        switch( sw ){
	case 1:
		printk(" ASPECT: 1:1");
                p->pesto[w].aspect = 100;        
		break;
	case 2:
		printk(" ASPECT: 4:3");
                p->pesto[w].aspect = 133;        
		break;
	case 3:
		printk(" ASPECT: 16:9");
                p->pesto[w].aspect = 177;        
		break;
	case 4:
		printk(" ASPECT: 2.21:1");
                p->pesto[w].aspect = 221;        
		break;

        case 5 ... 15:
		printk(" ASPECT: reserved");
                p->pesto[w].aspect = 0;        
		break;

        default:
                p->pesto[w].aspect = 0;        
                return -1;
	}

        printk("  Size = %dx%d",hsize,vsize);

        sw = (int)(headr[3]&0x0F);

        switch ( sw ) {
	case 1:
		printk("  FRAMERATE: 23.976 fps");
                p->pesto[w].framerate = 24000/1001;
		form = -1;
		break;
	case 2:
		printk("  FRAMERATE: 24 fps");
                p->pesto[w].framerate = 24;
		form = -1;
		break;
	case 3:
		printk("  FRAMERATE: 25 fps");
                p->pesto[w].framerate = 25;
		form = VIDEO_MODE_PAL;
		break;
	case 4:
		printk("  FRAMERATE: 29.97 fps");
                p->pesto[w].framerate = 30000/1001;
		form = VIDEO_MODE_NTSC;
		break;
	case 5:
		printk("  FRAMERATE: 30 fps");
                p->pesto[w].framerate = 30;
		form = VIDEO_MODE_NTSC;
		break;
	case 6:
		printk("  FRAMERATE: 50 fps");
                p->pesto[w].framerate = 50;
		form = VIDEO_MODE_PAL;
		break;
	case 7:
		printk("  FRAMERATE: 60 fps" );
                p->pesto[w].framerate = 60;
		form = VIDEO_MODE_NTSC;
		break;
	}

        if (c < MAX_AV-3){
                memcpy(headr,buf+c,3);
                
                bit_rate = 400*(((headr[0] << 10) & 0x0003FC00UL) 
                                | ((headr[1] << 2) & 0x000003FCUL) | 
                                    (((headr[2] & 0xC0) >> 6) & 0x00000003UL));
	
                printk("  BRate: %d.%d Mbit/s",bit_rate/1000000,
                       (bit_rate%1000000)/10000);
        }
	
        printk("\n");
        p->pesto[w].video_format = form;

        return form;
}


static int find_audio_format(trans_t *p, u8 *buf)
{
	u8 headr[2];
	int found=0;
        int c = 0;
	int layer;
        int w = p->which_pesto;
        int bit_rate,frequency;

	static unsigned int bitrates[3][16] =
	{{0,32,64,96,128,160,192,224,256,288,320,352,384,416,448,0},
	 {0,32,48,56,64,80,96,112,128,160,192,224,256,320,384,0},
	 {0,32,40,48,56,64,80,96,112,128,160,192,224,256,320,0}};

	static int freq[4] = {441, 480, 320, 0};

	while (c < MAX_AVA - 8 && !found) {
                memcpy(headr,buf+c,2);
                if (headr[0] == 0xFF && (headr[1] & 0xF8) == 0xF8
                        && headr[1] != 0xFF){
                        found = 1;
                        c+=2;
                }
		if (! found)  c++;
	}
        
        if (!found) return -1;
        
        
	layer = (headr[1] & 0x06) >> 1;


        printk("MPEG%d: audio: ",p->pesto[w].mpeg);
        memcpy(headr,buf+c,2);

        printk(" Layer: %d", layer);


	bit_rate = bitrates[(3-layer)][(headr[0] >> 4 )]*1000;

	if ( bit_rate == 0)
		printk("  Bit rate: free");
	else if (bit_rate == 0xf)
		printk("  BRate: reserved");
	else
		printk ("  BRate: %d kb/s", bit_rate/1000);
	

	frequency = (headr[0] & 0x0c ) >> 2;
	if ( frequency == 3)
		printk ("  Freq: reserved\n");
	else
		printk ("  Freq: %2d.%d kHz\n", 
			 freq[frequency]/10,(freq[frequency])%10);

        return frequency;
}
#endif

void init_p2t(p2t_t *p, dvb_demux_feed_t *feed)
{
	memset(p->pes,0,TS_SIZE);
	p->counter = 0;
	p->pos = 0;
	p->frags = 0;
	if (feed) p->feed = feed;
}

void clear_p2t(p2t_t *p)
{
	memset(p->pes,0,TS_SIZE);
	p->counter = 0;
	p->pos = 0;
	p->frags = 0;
}


long int find_pes_header(u8 const *buf, long int length, int *frags)
{
	int c = 0;
	int found = 0;

	*frags = 0;

	while (c < length-3 && !found) {
		if (buf[c] == 0x00 && buf[c+1] == 0x00 && 
		    buf[c+2] == 0x01) {
			switch ( buf[c+3] ) {
				
			case PROG_STREAM_MAP:
			case PRIVATE_STREAM2:
			case PROG_STREAM_DIR:
			case ECM_STREAM     :
			case EMM_STREAM     :
			case PADDING_STREAM :
			case DSM_CC_STREAM  :
			case ISO13522_STREAM:
			case PRIVATE_STREAM1:
			case AUDIO_STREAM_S ... AUDIO_STREAM_E:
			case VIDEO_STREAM_S ... VIDEO_STREAM_E:
				found = 1;
				break;
				
			default:
				c++;
				break;
			}	
		} else c++;
	}
	if (c == length-3 && !found){
		if (buf[length-1] == 0x00) *frags = 1;
		if (buf[length-2] == 0x00 &&
		    buf[length-1] == 0x00) *frags = 2;
		if (buf[length-3] == 0x00 &&
		    buf[length-2] == 0x00 &&
		    buf[length-1] == 0x01) *frags = 3;
		return -1;
	}

	return c;
}

void pes_to_ts( u8 const *buf, long int length, u16 pid, p2t_t *p)
{
	int c,c2,l,add;
	int check,rest;

	c = 0;
	c2 = 0;
	if (p->frags){
		check = 0;
		switch(p->frags){
		case 1:
			if ( buf[c] == 0x00 && buf[c+1] == 0x01 ){
				check = 1;
				c += 2;
			}
			break;
		case 2:
			if ( buf[c] == 0x01 ){
				check = 1;
				c++;
			}
			break;
		case 3:
			check = 1;
		}
		if(check){
			switch ( buf[c] ) {
				
			case PROG_STREAM_MAP:
			case PRIVATE_STREAM2:
			case PROG_STREAM_DIR:
			case ECM_STREAM     :
			case EMM_STREAM     :
			case PADDING_STREAM :
			case DSM_CC_STREAM  :
			case ISO13522_STREAM:
			case PRIVATE_STREAM1:
			case AUDIO_STREAM_S ... AUDIO_STREAM_E:
			case VIDEO_STREAM_S ... VIDEO_STREAM_E:
				p->pes[0] = 0x00;
				p->pes[1] = 0x00;
				p->pes[2] = 0x01;
				p->pes[3] = buf[c];
				p->pos=4;
				memcpy(p->pes+p->pos,buf+c,(TS_SIZE-4)-p->pos);
				c += (TS_SIZE-4)-p->pos;
				p_to_t(p->pes,(TS_SIZE-4),pid,&p->counter,
				       p->feed);
				clear_p2t(p);
				break;
				
			default:
				c=0;
				break;
			}
		}
		p->frags = 0;
	}
		
	if (p->pos){
		c2 = find_pes_header(buf+c,length-c,&p->frags);
		if (c2 >= 0 && c2 < (TS_SIZE-4)-p->pos){
			l = c2+c;
		} else l = (TS_SIZE-4)-p->pos;
		memcpy(p->pes+p->pos,buf,l);
		c += l;
		p->pos += l;
		p_to_t(p->pes,p->pos,pid,&p->counter, p->feed);
		clear_p2t(p);
	}
			
	add = 0;
	while (c < length){
		c2 = find_pes_header(buf+c+add,length-c-add,&p->frags);
		if (c2 >= 0) {
			c2 += c+add;
			if (c2 > c){
				p_to_t(buf+c,c2-c,pid,&p->counter,
				       p->feed);
				c = c2;
				clear_p2t(p);
				add = 0;
			} else add = 1;
		} else {
			l = length-c;
			rest = l % (TS_SIZE-4);
			l -= rest;
			p_to_t(buf+c,l,pid,&p->counter,
			       p->feed);
			memcpy(p->pes,buf+c+l,rest);
			p->pos = rest;
			c = length;
		}
	}
}


int write_ts_header2(u16 pid, u8 *counter, int pes_start, u8 *buf, u8 length)
{
	int i;
	int c = 0;
	int fill;
	u8 tshead[4] = { 0x47, 0x00, 0x00, 0x10}; 
        

	fill = (TS_SIZE-4)-length;
        if (pes_start) tshead[1] = 0x40;
	if (fill) tshead[3] = 0x30;
        tshead[1] |= (u8)((pid & 0x1F00) >> 8);
        tshead[2] |= (u8)(pid & 0x00FF);
        tshead[3] |= ((*counter)++ & 0x0F) ;
        memcpy(buf,tshead,4);
	c+=4;


	if (fill){
		buf[4] = fill-1;
		c++;
		if (fill >1){
			buf[5] = 0x00;
			c++;
		}
		for ( i = 6; i < fill+4; i++){
			buf[i] = 0xFF;
			c++;
		}
	}

        return c;
}


void p_to_t(u8 const *buf, long int length, u16 pid, u8 *counter, 
            dvb_demux_feed_t *feed)
{
  
	int l, pes_start;
	u8 obuf[TS_SIZE];
	long int c = 0;

	pes_start = 0;
	if ( length > 3 && 
             buf[0] == 0x00 && buf[1] == 0x00 && buf[2] == 0x01 )
		switch (buf[3]){
			case PROG_STREAM_MAP:
			case PRIVATE_STREAM2:
			case PROG_STREAM_DIR:
			case ECM_STREAM     :
			case EMM_STREAM     :
			case PADDING_STREAM :
			case DSM_CC_STREAM  :
			case ISO13522_STREAM:
			case PRIVATE_STREAM1:
			case AUDIO_STREAM_S ... AUDIO_STREAM_E:
			case VIDEO_STREAM_S ... VIDEO_STREAM_E:
				pes_start = 1;
				break;
				
			default:
				break;
		}			

	while ( c < length ){
		memset(obuf,0,TS_SIZE);
		if (length - c >= (TS_SIZE-4)){
			l = write_ts_header2(pid, counter, pes_start
					     , obuf, (TS_SIZE-4));
			memcpy(obuf+l, buf+c, TS_SIZE-l);
			c += TS_SIZE-l;
		} else { 
			l = write_ts_header2(pid, counter, pes_start
					     , obuf, length-c);
			memcpy(obuf+l, buf+c, TS_SIZE-l);
			c = length;
		}
                feed->cb.ts(obuf, 188, 0, 0, &feed->feed.ts, DMX_OK); 
		pes_start = 0;
	}
}


/****************************************************************************
 * V4L SECTION
 ****************************************************************************/

static struct video_channel dvb_channels[1] = {
        { 0, "DVB", 1, 3, 1, 1},
};

static int dvb_ioctl(struct video_device *dev, unsigned int cmd, void *arg)
{
	struct dvb_struct* dvb = (struct dvb_struct*)dev;
        
	switch (cmd)
	{
		/* Get capabilities */
        case VIDIOCGCAP:
        {
                struct video_capability b;
                
                dprintk(KERN_ERR "dvb: VIDIOCGCAP called\n");

                strcpy(b.name, &dev->name[0]);
			
                b.type 		= dvb->video.type;
						
                b.channels	= 1;	
                b.audios	= 2;
                b.maxwidth	= 768;
                b.maxheight	= 576;
                b.minwidth	= 32;
                b.minheight	= 32;
                
                if(copy_to_user(arg,&b,sizeof(b)))
                        return -EFAULT;
                return 0;
        }
		/* Get channel info (sources) */
        case VIDIOCGCHAN:
        {
                struct video_channel v;
                
                dprintk(KERN_ERR "dvb: VIDIOCGCHAN called\n");
                
                if(copy_from_user(&v, arg,sizeof(v)))
                        return -EFAULT;
                
                /* copy the desired structure to v */
                memcpy(&v, &dvb_channels[v.channel], sizeof(v));
                
                if(copy_to_user(arg,&v,sizeof(v)))
                        return -EFAULT;
                
                return 0;
                
        }
		/* Set channel 	*/
        case VIDIOCSCHAN:
        {
                struct video_channel v;
		if(copy_from_user(&v, arg,sizeof(v)))
			return -EFAULT;
		
                
		if (v.channel>0)
			return -EINVAL;
		if (v.norm > 1)
			return -EOPNOTSUPP;

                dvb->vidmode=v.norm;
                SetMode(dvb, v.norm);

		return 0;
        }
                
		/* Get tuner abilities */
        case VIDIOCGTUNER:
        {
                struct video_tuner v;
                
                dprintk(KERN_ERR "dvb: VIDIOCGTUNER called\n");
                
                if(copy_from_user(&v,arg,sizeof(v)))
                        return -EFAULT;
                
                /* only channel 0 has a tuner */	
                if(v.tuner != 0 )
                        return -EINVAL;
                
                /* fill the structure */
                strcpy(v.name, "DVB");
                v.rangelow  = 0x00000000;
                v.rangehigh = 0xffffffff;
                
                v.flags= VIDEO_TUNER_PAL | VIDEO_TUNER_NTSC;
                v.mode = dvb->vidmode;
		
                /* fixme: fill in signal strength here */
                v.signal = 0xffff;
                
                if(copy_to_user(arg,&v,sizeof(v)))
                        return -EFAULT;
                
                return 0;
        }
		/* Tune the tuner for the current channel */
        case VIDIOCSTUNER:
        {
                struct video_tuner v;
                
                dprintk(KERN_ERR "dvb: VIDIOCSTUNER called\n");
                
                if(copy_from_user(&v, arg, sizeof(v)))
                        return -EFAULT;
                
                /* only channel 0 has a tuner */
                if( v.tuner != 0)
                        return -EINVAL;
                
                /* check if format is supported */
                if(    v.mode != VIDEO_MODE_PAL 
                       && v.mode != VIDEO_MODE_NTSC
                    /* && v.mode != VIDEO_MODE_SECAM */ )
                       return -EOPNOTSUPP;
                
                dvb->vidmode=v.mode;
                SetMode(dvb, v.mode);
                
                return 0;
        }
		/* Get picture properties */
        case VIDIOCGPICT:
        {
                struct video_picture p;
                
                saacomm(SAA7146_V4L_GPICT, &p);
                
                if(copy_to_user(arg, &p, sizeof(struct video_picture)))
                        return -EFAULT;
                
                dprintk(KERN_ERR "dvb: VIDIOCGPICT called: b:%d c:%d s:%d d:%d p:%d\n",p.brightness, p.contrast, p.colour, p.depth, p.palette);
                
                return 0;
        }
		/* Set picture properties */
        case VIDIOCSPICT:
        {
                struct video_picture p;
                
                if(copy_from_user(&p, arg,sizeof(struct video_picture)))
                        return -EFAULT;
                
                dprintk(KERN_ERR "dvb: VIDIOCGPICT called: b:%d c:%d s:%d d:%d p:%d\n",p.brightness, p.contrast, p.colour, p.depth, p.palette);
                
                saacomm(SAA7146_V4L_SPICT, &p);
                
                return 0;
                
        }
		/* Set the video overlay window */
        case VIDIOCSWIN:
        {
                struct video_window w;
                
                if(copy_from_user(&w,arg,sizeof(struct video_window)))
                        return -EFAULT;
                
                dprintk(KERN_ERR "dvb: VIDIOCSWIN called: clips: %d, x:%d, y:%d, h:%d, w:%d\n", w.clipcount,w.x, w.y, w.height,w.width);
                
                saacomm(SAA7146_V4L_SWIN, &w);
                
                return 0;
        }
		/* Set the video overlay window - passes clip list for hardware smarts , chromakey etc */
        case VIDIOCGWIN:
        {
                dprintk(KERN_ERR "dvb: VIDIOCGWIN called\n");
                return 0;
        }
		/* Start, end capture */
        case VIDIOCCAPTURE:
        {
                int v;
                
                if(copy_from_user(&v, arg,sizeof(v)))
                        return -EFAULT;
                
                dprintk(KERN_ERR "dvb: VIDIOCCAPTURE called, mode:%d (0=disable)\n",v);
                
                saacomm(SAA7146_V4L_CCAPTURE, &v);
                
                return 0;
        }
		/* Get frame buffer */
        case VIDIOCGFBUF:
        {
                struct video_buffer b;
                
                dprintk(KERN_ERR "dvb: VIDIOCGFBUF called\n");
                
                saacomm(SAA7146_V4L_GFBUF, &b);
                
                if(copy_to_user(arg, &b,sizeof(struct video_buffer)))
                        return -EFAULT;
                
                return 0;
                
        }
		/* Set frame buffer - root only */
        case VIDIOCSFBUF:
        {
                struct video_buffer b;
                u32 vid = (vidmem << 16) | vidlow;
                
                if(copy_from_user(&b, arg,sizeof(struct video_buffer)))
                        return -EFAULT;
                
                /* see if vidmem-override is requested */
                if( 0 != vidmem ) {
                        printk(KERN_ERR "dvb: video-memory-override. (0x%08x)\n",vid);
                        b.base = (void*)vid;
                }
                
                saacomm(SAA7146_V4L_SFBUF, &b);
                
                dprintk(KERN_ERR "dvb: VIDIOCSFBUF called\n");
                
                return 0;		
        }
		/* Video key event - to dev 255 is to all - cuts capture on all DMA windows with this key (0xFFFFFFFF == all) */
        case VIDIOCKEY:
        {
                dprintk(KERN_ERR "dvb: VIDIOCKEY called\n");
                return 0;
        }
		/* Get tuner */
        case VIDIOCGFREQ:
        {
                unsigned long v = dvb->front.freq;
                
                dprintk(KERN_ERR "dvb: VIDIOCGFREQ called\n");
                
                if(copy_to_user(arg,&v,sizeof(v)))
                        return -EFAULT;
        }
		/* Set tuner */
        case VIDIOCSFREQ:
        {
                int v;
                
                if(copy_from_user(&v, arg, sizeof(v)))
                        return -EFAULT;
                
                //dprintk(KERN_ERR "dvb: VIDIOCSFREQ called with freq %u; audio:%d\n", v, dvb->current_audio_state);
                
                /* mute audio to avoid strange tuning noises */	
                
                //mdelay( TUNER_NOISE_DELAY );
                
                /* tune in desired frequency */			
                //tuner_command(dvb, TUNER_SET_TVFREQ,&v); 
                //mdelay( TUNER_NOISE_DELAY );
                
                /* restore audio if it should not be muted */
                
                dvb->front.freq = v;
                return 0;
        }
		/* Get audio info */	
        case VIDIOCGAUDIO:
        {
                struct video_audio v;
                
                v.flags = VIDEO_AUDIO_MUTABLE;
                /* let's auto-detect */
                return 0;
        }
		/* Audio source, mute etc */
        case VIDIOCSAUDIO:
        {
          //struct video_audio v;
                
                
                return 0;
        }
                /* Sync with mmap grabbing */
        case VIDIOCSYNC:
        {
                int i = 0, frame = 0;
                
                if(copy_from_user((void *)&frame,arg,sizeof(int)))
                        return -EFAULT;
                
                dprintk(KERN_ERR "dvb: VIDIOCSYNC called - frame: %d\n",frame);
                
                /* simply pass the requested frame-number to the corresponding
                   saa7146-function ... */
                i = saacomm(SAA7146_V4L_CSYNC, &frame);
                
                return i;
        }
		/* Grab frames */
        case VIDIOCMCAPTURE:
        {
                struct video_mmap vm;
                int i = 0;
                
                if(copy_from_user((void *)&vm, (void *) arg, sizeof(vm)))
                        return -EFAULT;
                
                dprintk(KERN_ERR "dvb: VIDIOCMCAPTURE called: fr:%d,"
                        "fmt:%d, w:%d, h:%d\n",
                        vm.frame, vm.format,vm.width, vm.height);
                
                /* simply pass the structure for the requested frame-number
                   to the corresponding saa7146-function ... */
                i = saacomm(SAA7146_V4L_CMCAPTURE, &vm);
                
                return i;
        }
                
		/* Memory map buffer info */
        case VIDIOCGMBUF:
        {
                struct video_mbuf mbuf;
                
                dprintk(KERN_ERR "dvb: VIDIOCGMBUF called\n");
                
                saacomm(SAA7146_V4L_GMBUF, &mbuf);
                
                if (copy_to_user((void *)arg, &mbuf, sizeof(struct video_mbuf)))
                        return -EFAULT;
                
                return 0;
                
        }
		/* Get attached units */
        case VIDIOCGUNIT:
        {
                struct video_unit vu;
                dprintk(KERN_ERR "dvb: VIDIOCGUNIT called\n");
                
                if(copy_to_user((void *)arg, (void *)&vu, sizeof(vu)))
                        return -EFAULT;
                return 0;
        }
		
        case VIDIOCGCAPTURE:
        {
                dprintk(KERN_ERR "dvb: VIDIOCGCAPTURE called\n");
                return 0;
        }
        case VIDIOCSCAPTURE:
        {
                dprintk(KERN_ERR "dvb: VIDIOCSCAPTURE called\n");
                return 0;
        }

        default:
                return -ENOIOCTLCMD;
	}
        return 0;
}

static int dvb_init_done(struct video_device *dev)
{
	fprintk(KERN_ERR "dvb: dvb_init_done called\n");
	return 0;
}

static int dvb_mmap(struct video_device *dev, const char *adr, unsigned long size)
{
	struct dvb_struct* dvb = (struct dvb_struct*)dev;
	struct saa7146_mmap_struct m;
	
	fprintk(KERN_ERR "dvb: dvb_mmap called, adr:%p, size:0x%08x\n",adr,(u32)size);
        
	m.adr = adr;
	m.size = size;
        
	if ( 0 != saacomm(SAA7146_DO_MMAP, &m) ) {
		printk(KERN_ERR "dvb: dvb_mmap failed!\n");
		return -1;
	}
        
	return 0;
}

static int dvb_open(struct video_device *dev, int flags)
{
	MOD_INC_USE_COUNT;
	return 0;   
}

static void dvb_close(struct video_device *dev)
{
	fprintk(KERN_ERR "dvb: dvb_close called\n");
        MOD_DEC_USE_COUNT;  
}

static unsigned int dvb_poll(struct dvb_struct *dvb, struct file *file,
	poll_table *wait)
{
        unsigned int mask=0;

        if (dvb->playing) {
                if (FREE_COND)
                        return (POLLOUT | POLLWRNORM);

                poll_wait(file, &dvb->avout.queue, wait);
          
                if (FREE_COND)
                        mask |= (POLLOUT | POLLWRNORM);
        } else /* if not playing: may play if asked for */
                mask = (POLLOUT | POLLWRNORM);
                
        return mask;
}

static unsigned int dvb_apoll(struct dvb_struct *dvb, struct file *file,
	poll_table *wait)
{
        unsigned int mask=0;

        if (dvb->playing) {
                if (dvb_buf_free(&dvb->aout)>20*1024)
                        return (POLLOUT | POLLWRNORM);

                poll_wait(file, &dvb->aout.queue, wait);
          
                if (dvb_buf_free(&dvb->aout)>20*1024)
                        mask |= (POLLOUT | POLLWRNORM);
        } else /* if not playing: may play if asked for */
                mask = (POLLOUT | POLLWRNORM);
                
        return mask;
}

/* template for video_device-structure */
static struct video_device dvb_template = {
name:	        "DVB Board",
type:	        VID_TYPE_TUNER		|
                VID_TYPE_CAPTURE	|
                VID_TYPE_OVERLAY	|
                VID_TYPE_CLIPPING	|
                VID_TYPE_FRAMERAM	|
                VID_TYPE_SCALES,
hardware:       VID_HARDWARE_SAA7146,
open:           dvb_open,
close:          dvb_close,
read:           NULL,
write:	        NULL,		
poll:	        NULL,
ioctl: 	        dvb_ioctl,
mmap:	        dvb_mmap,
initialize:	dvb_init_done,
priv:   	NULL,
busy:           0,
minor:          0
};

static int vid_register(struct dvb_struct *dvb)
{
        /* Copy structure */
        memcpy( &dvb->video, &dvb_template, sizeof(struct video_device));

#if LINUX_VERSION_CODE < 0x020405
        if( 0 > (video_register_device(&dvb->video, VFL_TYPE_GRABBER))) {
#else
        if( 0 > (video_register_device(&dvb->video, VFL_TYPE_GRABBER, -1))) {
#endif
                printk(KERN_ERR "dvb: can't register videodevice\n");
                return -1;
        }
        return 0;
}

static inline int vid_unregister(struct dvb_struct *dvb)
{
        if(dvb->video.minor!=-1)
                video_unregister_device(&dvb->video);
        return 0;
}

/****************************************************************************
 * END OF V4L SECTION
 ****************************************************************************/


/****************************************************************************
 * DVB API SECTION
 ****************************************************************************/


/******************************************************************************
 * hardware filter functions
 ******************************************************************************/

static int 
StartHWFilter(dvb_demux_filter_t *dvbdmxfilter)
{
        dvb_demux_feed_t *dvbdmxfeed=dvbdmxfilter->feed;
        struct dvb_struct *dvb=
                (struct dvb_struct *) dvbdmxfeed->demux->priv;
        u16 buf[20];
        int ret, i;
        u16 handle;
        u16 mode=0x0320;
        
        if (dvbdmxfilter->type==DMX_TYPE_SEC) {
                buf[4]=(dvbdmxfilter->filter.filter_value[0]<<8)|
                        dvbdmxfilter->filter.filter_mask[0];
                for (i=3; i<18; i++)
                        buf[i+4-2]=(dvbdmxfilter->filter.filter_value[i]<<8)|
                                    dvbdmxfilter->filter.filter_mask[i];
                mode=4;
        } else
        if ((dvbdmxfeed->ts_type & TS_PACKET) &&
            !(dvbdmxfeed->ts_type & TS_PAYLOAD_ONLY)) 
                init_p2t(&dvb->p2t_filter[dvbdmxfilter->index], dvbdmxfeed);

        buf[0] = (COMTYPE_PID_FILTER << 8) + AddPIDFilter; 
        buf[1] = 16;	
        buf[2] = dvbdmxfeed->pid;
        buf[3] = mode;

        ret=CommandRequest(dvb, buf, 20, &handle, 1);          
        if (ret<0)
                return ret;

        dvb->handle2filter[handle]=dvbdmxfilter;
        dvbdmxfilter->hw_handle=handle;

        return ret;
}

static int 
StopHWFilter(dvb_demux_filter_t *dvbdmxfilter)
{
        struct dvb_struct *dvb=
          (struct dvb_struct *) dvbdmxfilter->feed->demux->priv;
        u16 buf[3];
        u16 answ[2];
        int ret;
        u16 handle;
                
        handle=dvbdmxfilter->hw_handle;
        if (handle>32) {
                dprintk("dvb: StopHWFilter tried to stop invalid filter %d.\n",
                       handle);
                dprintk("dvb: filter type = %d\n",  dvbdmxfilter->type);
                return 0;
        }

        dvb->handle2filter[handle]=NULL;

        buf[0] = (COMTYPE_PID_FILTER << 8) + DelPIDFilter; 
        buf[1] = 1;	
        buf[2] = handle;
        ret=CommandRequest(dvb, buf, 3, answ, 2);          

        if (answ[1] != handle) {
                dprintk("dvb: filter %d shutdown error :%d\n", handle, answ[1]);
                ret=-1;
        }
        return ret;
}

static int 
dvb_write_to_decoder(dvb_demux_feed_t *dvbdmxfeed, u8 *buf, size_t count)
{
        dvb_demux_t *dvbdmx=dvbdmxfeed->demux;
        struct dvb_struct *dvb=(struct dvb_struct *) dvbdmx->priv;
        ipack *ipack=&dvb->ipack[dvbdmxfeed->pes_type];

        if (dvbdmxfeed->pes_type>1)
                return -1;
        if (!(buf[3]&0x10)) // no payload?
                return -1;

        if (buf[1]&0x40) 
                send_ipack_rest(ipack);

        if (buf[3]&0x20) {  // adaptation field?
                count-=buf[4]+1;
                buf+=buf[4]+1;
                if (!count)
                        return 0;
        }
        
        instant_repack(buf+4, count-4, &dvb->ipack[dvbdmxfeed->pes_type]);
        return 0;
}

static void
dvb_feed_start_pid(dvb_demux_feed_t *dvbdmxfeed)
{
        dvb_demux_t *dvbdmx=dvbdmxfeed->demux;
        struct dvb_struct *dvb=(struct dvb_struct *) dvbdmx->priv;
        u16 *pid=dvbdmx->pids, npids[5];
        int i;
        
        npids[0]=npids[1]=npids[2]=npids[3]=0xffff;
        npids[4]=0xffff;
        i=dvbdmxfeed->pes_type;
        npids[i]=(pid[i]&0x8000) ? 0 : pid[i];
        if ((i==2) && npids[i] && (dvbdmxfeed->ts_type & TS_PACKET)) {
                npids[i]=0;
                ChangePIDs(dvb, npids[1], npids[0], npids[2], npids[3], npids[4]);
                StartHWFilter(dvbdmxfeed->filter); 
                return;
        }
        if (dvbdmxfeed->pes_type<=2)
                ChangePIDs(dvb, npids[1], npids[0], npids[2], npids[3], npids[4]);
        if (dvbdmxfeed->pes_type<2 && npids[0])
                if (dvb->mon_tuning==3)
                        outcom(dvb, COMTYPE_PIDFILTER, Scan, 0);
        if ((dvbdmxfeed->ts_type & TS_PACKET)) { 
                if (dvbdmxfeed->pes_type == 0 && 
                    !(dvbdmx->pids[0]&0x8000))
                        AV_StartRecord(dvb, RP_AUDIO, 
                                       dvbdmxfeed);
                if (dvbdmxfeed->pes_type == 1 && 
                    !(dvbdmx->pids[1]&0x8000))
                        AV_StartRecord(dvb, RP_VIDEO, 
                                       dvbdmxfeed);
        }
}

static void
dvb_feed_stop_pid(dvb_demux_feed_t *dvbdmxfeed)
{
        dvb_demux_t *dvbdmx=dvbdmxfeed->demux;
        struct dvb_struct *dvb=(struct dvb_struct *) dvbdmx->priv;
        u16 *pid=dvbdmx->pids, npids[5];
        int i;
        
        if (dvbdmxfeed->pes_type<=1) {
                AV_Stop(dvb, dvbdmxfeed->pes_type ? 
                        RP_VIDEO : RP_AUDIO);
                if (!dvb->rec_mode)
                        dvbdmx->recording=0;
                if (!dvb->playing)
                        dvbdmx->playing=0;
        }
        npids[0]=npids[1]=npids[2]=npids[3]=0xffff;
        npids[4]=0xffff;
        i=dvbdmxfeed->pes_type;
        switch (i) {
        case 2: //teletext
                if (dvbdmxfeed->ts_type & TS_PACKET)
                        StopHWFilter(dvbdmxfeed->filter); 
                npids[2]=0;
                break;
        case 0:
        case 1:
        case 4:
                if (!pids_off) 
                        return;
                npids[i]=(pid[i]&0x8000) ? 0 : pid[i];
                break;
        }
        ChangePIDs(dvb, npids[1], npids[0], npids[2], npids[3], npids[4]);
}

static int 
dvb_start_feed(dvb_demux_feed_t *dvbdmxfeed)
{
        dvb_demux_t *dvbdmx=dvbdmxfeed->demux;
        struct dvb_struct *dvb=(struct dvb_struct *) dvbdmx->priv;

        if (!dvbdmx->dmx.frontend)
                return -EINVAL;

        if (dvb->card_type==DVB_CARD_TT_BUDGET)
                return TTBStart(dvb); 

        if (dvbdmxfeed->type == DMX_TYPE_TS) {
	        if ((dvbdmxfeed->ts_type & TS_DECODER) 
		    && (dvbdmxfeed->pes_type<DMX_TS_PES_OTHER)) {
		        switch (dvbdmx->dmx.frontend->source) {
			case DMX_MEMORY_FE: 
			        if (dvbdmxfeed->ts_type & TS_DECODER)
				       if (dvbdmxfeed->pes_type<2 && 
                                           !(dvbdmx->pids[0]&0x8000) &&
					   !(dvbdmx->pids[1]&0x8000)) {
                                               dvb_buf_flush(&dvb->avout);
                                               dvb_buf_flush(&dvb->aout);
                                               AV_StartPlay(dvb,RP_AV);
                                               dvbdmx->playing=1;
					}
				break;
			default:
                                dvb_feed_start_pid(dvbdmxfeed);
				break;
			}
		} else
		        if ((dvbdmxfeed->ts_type & TS_PACKET) &&
                            (dvbdmx->dmx.frontend->source!=DMX_MEMORY_FE))
                                StartHWFilter(dvbdmxfeed->filter); 
        }
        
        if (dvbdmxfeed->type == DMX_TYPE_SEC) {
                int i;

	        for (i=0; i<dvbdmx->filternum; i++) {
		        if (dvbdmx->filter[i].state!=DMX_STATE_READY)
			        continue;
			if (dvbdmx->filter[i].type!=DMX_TYPE_SEC)
			        continue;
			if (dvbdmx->filter[i].filter.parent!=&dvbdmxfeed->feed.sec)
			        continue;

			dvbdmxfeed->feed.sec.is_filtering=1;
			dvbdmx->filter[i].state=DMX_STATE_GO;
                        if (dvbdmx->dmx.frontend->source!=DMX_MEMORY_FE)
                                StartHWFilter(&dvbdmx->filter[i]); 
                }
	}
        return 0;
}

static int 
dvb_stop_feed(dvb_demux_feed_t *dvbdmxfeed)
{
        dvb_demux_t *dvbdmx=dvbdmxfeed->demux;
        struct dvb_struct *dvb=(struct dvb_struct *) dvbdmx->priv;

        if (dvb->card_type==DVB_CARD_TT_BUDGET)
                return TTBStop(dvb); 

        if (dvbdmxfeed->type == DMX_TYPE_TS) {
                if (dvbdmxfeed->ts_type & TS_DECODER) {
                        if (dvbdmxfeed->pes_type>=DMX_TS_PES_OTHER)
                                return -EINVAL;
                        if (!dvbdmx->pesfilter[dvbdmxfeed->pes_type])
                                return -EINVAL;
                        dvbdmx->pids[dvbdmxfeed->pes_type]|=0x8000;
                        dvbdmx->pesfilter[dvbdmxfeed->pes_type]=0;
                }
                if (dvbdmxfeed->ts_type & TS_DECODER &&
		    (dvbdmxfeed->pes_type<DMX_TS_PES_OTHER)) {
                        dvb_feed_stop_pid(dvbdmxfeed);
                } else 
                        if ((dvbdmxfeed->ts_type & TS_PACKET) &&
                            (dvbdmx->dmx.frontend->source!=DMX_MEMORY_FE))
                                StopHWFilter(dvbdmxfeed->filter); 
        }
        
        if (dvbdmxfeed->type == DMX_TYPE_SEC) {
                int i;

	        for (i=0; i<dvbdmx->filternum; i++)
		        if (dvbdmx->filter[i].state==DMX_STATE_GO && 
			    dvbdmx->filter[i].filter.parent==&dvbdmxfeed->feed.sec) {
			        dvbdmx->filter[i].state=DMX_STATE_READY;
                                if (dvbdmx->dmx.frontend->source!=DMX_MEMORY_FE)
                                        StopHWFilter(&dvbdmx->filter[i]); 
                }
	}

        return 0;
}


/******************************************************************************
 * SEC device file operations
 ******************************************************************************/

static void add_qpsk_event(qpsk_t *qpsk, struct qpskEvent *ev)
{
        int wp;
        struct timeval tv;

        get_fast_time(&tv);
        ev->timestamp=tv.tv_sec;
        
        spin_lock(&qpsk->eventlock);
        wp=qpsk->eventw;
        wp=(wp+1)%QPSK_MAX_EVENT;
        if (wp==qpsk->eventr) {
                qpsk->overflow=1;
                qpsk->eventr=(qpsk->eventr+1)%QPSK_MAX_EVENT;
        }
        memcpy(&qpsk->events[qpsk->eventw], ev, sizeof(struct qpskEvent));
        qpsk->eventw=wp;
        spin_unlock(&qpsk->eventlock);
        wake_up(&qpsk->eventq);
}

static void add_qam_event(qam_t *qam, struct qamEvent *ev)
{
        int wp;
        struct timeval tv;

        dprintk("add_qam_event\n");
        get_fast_time(&tv);
        ev->timestamp=tv.tv_sec;
        
        spin_lock(&qam->eventlock);
        wp=qam->eventw;
        wp=(wp+1)%QAM_MAX_EVENT;
        if (wp==qam->eventr) {
                qam->overflow=1;
                qam->eventr=(qam->eventr+1)%QAM_MAX_EVENT;
        }
        memcpy(&qam->events[qam->eventw], ev, sizeof(struct qamEvent));
        qam->eventw=wp;
        spin_unlock(&qam->eventlock);
        wake_up(&qam->eventq);
}

int secSetTone(struct dvb_struct *dvb, secToneMode mode)
{
        int val;

        switch (mode) {
        case SEC_TONE_ON:
                val=1;
                break;
        case SEC_TONE_OFF:
                val=0;
                break;
        default:
          return -EINVAL;
        }
        dvb->front.ttk=val;
        Set22K(dvb, val);
        return 0;
}

int secSetVoltage(struct dvb_struct *dvb, secVoltage voltage)
{
        int power=1, volt=0;
                       
        switch (voltage) {
        case SEC_VOLTAGE_LT: //FIXME: Test which one is true
        case SEC_VOLTAGE_OFF:
                power=0;
                break;
        case SEC_VOLTAGE_13:
                volt=0;
                break;
        case SEC_VOLTAGE_18:
                volt=1;
                break;
        case SEC_VOLTAGE_13_5:
        case SEC_VOLTAGE_18_5:
        default:
                return -EINVAL;
        }
        dvb->front.volt=volt;
        // this guarantees that running SEC commands are done atomically
        Set22K(dvb, dvb->front.ttk);
        SetVoltage(dvb, power, volt);
        return 0;
}

int secSendSequence(struct dvb_struct *dvb, struct secCmdSequence *seq)
{
        int i, ret, burst, len;
        struct secCommand scommands;
        u8 msg[16];

        switch (seq->miniCommand) {
        case SEC_MINI_NONE:
                burst=-1;
                break;
        case SEC_MINI_A:
                burst=0;
                break;
        case SEC_MINI_B:
                burst=1;
                break;
        default:
                return -EINVAL;
        }

        for (i=0; i<seq->numCommands; i++) {
                if (copy_from_user(&scommands, &seq->commands[i], 
                                   sizeof(struct secCommand)))
                        continue;
                switch (scommands.type) {
                case SEC_CMDTYPE_DISEQC:
                        len=scommands.u.diseqc.numParams;
                        if (len>SEC_MAX_DISEQC_PARAMS)
                                return -EINVAL;

                        msg[0]=0xe0;
                        msg[1]=scommands.u.diseqc.addr;
                        msg[2]=scommands.u.diseqc.cmd;
                        memcpy(msg+3, &scommands.u.diseqc.params, len);
                        SendDiSEqCMsg(dvb, len+3, msg, -1);
                        /* repeat for the case of cascaded (and/or hearing 
                           impared) switches */
                        msg[0]=0xe1;
                        SendDiSEqCMsg(dvb, len+3, msg, -1);
                        break;
                case SEC_CMDTYPE_PAUSE:
                        //FIXME: fix firmware to allow programmed pause length
                        SendDiSEqCMsg(dvb, 0, msg, -1);
                        break;
                case SEC_CMDTYPE_VSEC:
                default:
                        return -EINVAL;
                }
        }
        
        if (burst!=-1)
                SendDiSEqCMsg(dvb, 0, msg, burst);
        
        ret=secSetVoltage(dvb, seq->voltage);
        if (ret<0)
                return ret;
        return secSetTone(dvb, seq->continuousTone);
}



/******************************************************************************
 * DVB device file operations
 ******************************************************************************/

#define INFU 32768

static dvb_devs_t dvbs_devs = {
        10,

        { 
                DVB_DEVICE_VIDEO_0, DVB_DEVICE_AUDIO_0, 
                DVB_DEVICE_SEC_0,   DVB_DEVICE_QPSKFE_0, 
                DVB_DEVICE_DEMUX_0, DVB_DEVICE_DVR_0, 
                DVB_DEVICE_CA_0,    -1,
                DVB_DEVICE_OSD_0,   DVB_DEVICE_NET_0
        },
        { INFU, INFU, INFU, INFU, INFU, 1, 1, 0, 1, INFU },
        { 1, 1, 1, 1, INFU, 1, 1, 0, 1, INFU }
};

static dvb_devs_t dvbc_devs = {
        10,
        { 
                DVB_DEVICE_VIDEO_0, DVB_DEVICE_AUDIO_0, 
                -1, -1, 
                DVB_DEVICE_DEMUX_0, DVB_DEVICE_DVR_0, 
                DVB_DEVICE_CA_0,    DVB_DEVICE_QAMFE_0,
                DVB_DEVICE_OSD_0,   DVB_DEVICE_NET_0
        },
        { INFU, INFU, 0, 0, INFU, 1, 1, INFU, 1, INFU },
        { 1, 1, 0, 0, INFU, 1, 1, 1, 1, INFU }
};

static dvb_devs_t ttbs_devs = {
        10,
        { 
                -1, -1, 
                DVB_DEVICE_SEC_0,   DVB_DEVICE_QPSKFE_0, 
                DVB_DEVICE_DEMUX_0, DVB_DEVICE_DVR_0, 
                -1,                 -1,
                -1,                 DVB_DEVICE_NET_0
        },
        { 0, 0, INFU, INFU, INFU, 1, 0, 0, 0, INFU },
        { 0, 0, 1, 1, INFU, 1, 0, 0, 0, INFU }
};

static dvb_devs_t ttbc_devs = {
        10,
        {
                -1, -1,
                -1 ,-1,
                DVB_DEVICE_DEMUX_0, DVB_DEVICE_DVR_0,
                -1,                 DVB_DEVICE_QAMFE_0
                -1,                 DVB_DEVICE_NET_0
        },
        { 0, 0, 0, 0, INFU, 1, 0, INFU, 0, INFU },
        { 0, 0, 0, 0, INFU, 1, 0, 1, 0, INFU }
};


static inline int
num2type(struct dvb_struct *dvb, int num)
{
        if (!dvb->dvb_devs)
                return -2;
        if (num>=dvb->dvb_devs->num)
                return -2;
        return dvb->dvb_devs->tab[num];
}


static int 
dvbdev_open(struct dvb_device *dvbdev, int num, 
            struct inode *inode, struct file *file)
{
        struct dvb_struct *dvb=(struct dvb_struct *) dvbdev->priv;
        int type=num2type(dvb, num);
        int ret=0;

        if (type<0)
                return -EINVAL;

        if (dvb->users[num] >= dvb->dvb_devs->max_users[num])
                return -EBUSY;

	if ((file->f_flags&O_ACCMODE)!=O_RDONLY) 
                if (dvb->writers[num] >= dvb->dvb_devs->max_writers[num])
		        return -EBUSY;

	switch (type) {
	case DVB_DEVICE_VIDEO_0:
                dvb_buf_flush(&dvb->aout);
                dvb_buf_flush(&dvb->avout);
                dvb->video_blank=true;
                dvb->audiostate.AVSyncState=true;
                dvb->videostate.streamSource=VIDEO_SOURCE_DEMUX;
                break;
	case DVB_DEVICE_AUDIO_0:
                dvb_buf_flush(&dvb->aout);
                dvb->audiostate.streamSource=AUDIO_SOURCE_DEMUX;
                break;
	case DVB_DEVICE_SEC_0:
                if (file->f_flags&O_NONBLOCK) 
                        ret=-EWOULDBLOCK;
                break;
	case DVB_DEVICE_QPSKFE_0:
                if (dvb->dvbtype != DVB_S)
                        return -EINVAL;
                break;
	case DVB_DEVICE_QAMFE_0:
                if (dvb->dvbtype != DVB_C)
                        return -EINVAL;
                break;
	case DVB_DEVICE_DEMUX_0:
                if ((file->f_flags&O_ACCMODE)!=O_RDWR)
                        return -EINVAL;
                ret=DmxDevFilterAlloc(&dvb->dmxdev, file);
                break;
	case DVB_DEVICE_DVR_0:
                ret=DmxDevDVROpen(&dvb->dmxdev, file);
                break;
	case DVB_DEVICE_CA_0:
                printk("ca_open\n");
                break;
	case DVB_DEVICE_OSD_0:
                break;
	case DVB_DEVICE_NET_0:
                break;
        default:
                return -EINVAL;
	}
	if (ret<0) 
	        return ret;
	if ((file->f_flags&O_ACCMODE)!=O_RDONLY)
		dvb->writers[num]++;
        dvb->users[num]++;
	MOD_INC_USE_COUNT;
        return ret;
}

static int 
dvbdev_close(struct dvb_device *dvbdev, int num, 
             struct inode *inode, struct file *file)
{
        struct dvb_struct *dvb=(struct dvb_struct *) dvbdev->priv;
        int type=num2type(dvb, num);
        int ret=0;

        if (type<0)
                return -EINVAL;

	switch (type) {
	case DVB_DEVICE_VIDEO_0:
	        AV_Stop(dvb, RP_VIDEO);
		break;
	case DVB_DEVICE_AUDIO_0:
	        AV_Stop(dvb, RP_AUDIO);
                break;
	case DVB_DEVICE_SEC_0:
                break;
	case DVB_DEVICE_QPSKFE_0:
                if (dvb->dvbtype != DVB_S)
                        return -EINVAL;
                break;
	case DVB_DEVICE_QAMFE_0:
                if (dvb->dvbtype != DVB_C)
                        return -EINVAL;
                break;
	case DVB_DEVICE_DEMUX_0:
                ret=DmxDevFilterFree(&dvb->dmxdev, file);
                break;
	case DVB_DEVICE_DVR_0:
                ret=DmxDevDVRClose(&dvb->dmxdev, file);
                break;
	case DVB_DEVICE_CA_0:
                break;
	case DVB_DEVICE_OSD_0:
                break;
	case DVB_DEVICE_NET_0:
                break;
        default:
                return -EINVAL;
	}
        /*
	if (ret<0) 
	        return ret;
        */
	if ((file->f_flags&O_ACCMODE)!=O_RDONLY)
		dvb->writers[num]--;
        dvb->users[num]--;
	MOD_DEC_USE_COUNT;
        return ret;
}

static ssize_t 
dvbdev_read(struct dvb_device *dvbdev, int num, 
            struct file *file, char *buf, size_t count, loff_t *ppos)
{
        struct dvb_struct *dvb=(struct dvb_struct *) dvbdev->priv;
        int type=num2type(dvb, num);

	switch (type) {
	case DVB_DEVICE_VIDEO_0:
		break;
	case DVB_DEVICE_AUDIO_0:
                break;
	case DVB_DEVICE_DEMUX_0:
                return DmxDevRead(&dvb->dmxdev, file, buf, count, ppos);
	case DVB_DEVICE_DVR_0:
                return DmxDevDVRRead(&dvb->dmxdev, file, buf, count, ppos);
	case DVB_DEVICE_CA_0:
                break;
	default:
	        return -EOPNOTSUPP;
	}
        return 0;
}

static ssize_t 
dvbdev_write(struct dvb_device *dvbdev, int num,
             struct file *file, 
             const char *buf, size_t count, loff_t *ppos)
{
        struct dvb_struct *dvb=(struct dvb_struct *) dvbdev->priv;
        int type=num2type(dvb, num);

	switch (type) {
	case DVB_DEVICE_VIDEO_0:
                if (dvb->videostate.streamSource!=VIDEO_SOURCE_MEMORY)
                        return -EPERM;
                return dvb_play(dvb, buf, count, file->f_flags&O_NONBLOCK, 1, 1);

	case DVB_DEVICE_AUDIO_0:
                if (dvb->audiostate.streamSource!=AUDIO_SOURCE_MEMORY)
                        return -EPERM;
                return dvb_aplay(dvb, buf, count, file->f_flags&O_NONBLOCK, 0);

	case DVB_DEVICE_DVR_0:
                return DmxDevDVRWrite(&dvb->dmxdev, file, buf, count, ppos);

	case DVB_DEVICE_CA_0:
                break;

	default:
	        return -EOPNOTSUPP;
	}
        return 0;
}



int qam_tune(struct dvb_struct *dvb, struct qamParameters *para)
{
        if (para->FEC_inner>7)
                return -EINVAL;
        
        dvb->new_srate=para->SymbolRate;
        dvb->new_freq=para->Frequency;
        dvb->new_fec=fectab[para->FEC_inner];
        dvb->new_qam=para->QAM;

        mon_do_tune(dvb);
        return 0;
}
       
int qpsk_tune(struct dvb_struct *dvb, struct qpskParameters *para)
{
        if (para->FEC_inner>7)
                return -EINVAL;
        
        dvb->new_srate=para->SymbolRate;
        dvb->new_freq=para->iFrequency*1000;
        dvb->new_fec=fectab[para->FEC_inner];
        
        mon_do_tune(dvb);
        return 0;
}


u8 iframe_header[] = { 0x00, 0x00, 0x01, 0xe0, 0x00, 0x00, 0x80, 0x00, 0x00 };

#define MIN_IFRAME 400000

void play_iframe(struct dvb_struct *dvb, u8 *buf, unsigned int len, int nonblock)
{
        int i,n;

        n=MIN_IFRAME/len+1;

        dvb_play(dvb, iframe_header, sizeof(iframe_header), 0, 1, 0);
        for (i=0; i<n; i++) 
                dvb_play(dvb, buf, len, 0, 1, 1);
        send_ipack_rest(&dvb->ipack[1]);
}


// FIXME: fill in table
uint32_t nest2snr[256] = {
};

static int 
dvbdev_ioctl(struct dvb_device *dvbdev, int num, 
             struct file *file, unsigned int cmd, unsigned long arg)
{
        struct dvb_struct *dvb=(struct dvb_struct *) dvbdev->priv;
        void *parg=(void *)arg;
        int type=num2type(dvb, num);

	switch (type) {
	case DVB_DEVICE_VIDEO_0:
                if (((file->f_flags&O_ACCMODE)==O_RDONLY) &&
                    (cmd!=VIDEO_GET_STATUS))
                        return -EPERM;

                switch (cmd) {

                case VIDEO_STOP:
		        dvb->videostate.playState=VIDEO_STOPPED;
                        if (dvb->videostate.streamSource==VIDEO_SOURCE_MEMORY)
                                AV_Stop(dvb, RP_VIDEO);
                        else
                                vidcom(dvb, 0x000e, 
                                       dvb->videostate.videoBlank ? 0 : 1);
                        dvb->trickmode=TRICK_NONE;
                        return 0; 

                case VIDEO_PLAY:
                        dvb->trickmode=TRICK_NONE;
		        if (dvb->videostate.playState==VIDEO_FREEZED) {
			        dvb->videostate.playState=VIDEO_PLAYING;
				vidcom(dvb, 0x000d, 0);
			} 

                        if (dvb->videostate.streamSource==VIDEO_SOURCE_MEMORY) {
                                if (dvb->playing==RP_AV) {
                                        outcom(dvb, COMTYPE_REC_PLAY, __Stop, 0);
                                        dvb->playing&=~RP_VIDEO;
                                }
                                AV_StartPlay(dvb,RP_VIDEO);
                                vidcom(dvb, 0x000d, 0);
                        } else {
                                AV_Stop(dvb, RP_VIDEO);
                                vidcom(dvb, 0x000d, 0);
                        }
		        dvb->videostate.playState=VIDEO_PLAYING;
                        break;

                case VIDEO_FREEZE:
		        dvb->videostate.playState=VIDEO_FREEZED;
                        if (dvb->videostate.streamSource==VIDEO_SOURCE_MEMORY)
                                //outcom(dvb, COMTYPE_REC_PLAY, __Pause, 0);
                                vidcom(dvb, 0x000e, 1);
                        else
                                vidcom(dvb, 0x0102, 1);
                        dvb->trickmode=TRICK_FREEZE;
                        break;

                case VIDEO_CONTINUE:
                        if (dvb->videostate.streamSource==VIDEO_SOURCE_MEMORY) 
                                //dvb->videostate.playState==VIDEO_FREEZED) 
                                outcom(dvb, COMTYPE_REC_PLAY, __Continue, 0);
                        vidcom(dvb, 0x000d, 0);
                        dvb->videostate.playState=VIDEO_PLAYING;
                        dvb->trickmode=TRICK_NONE;
                       break;

                case VIDEO_SELECT_SOURCE:
                        dvb->videostate.streamSource=(videoStreamSource_t) arg;
                        break;

                case VIDEO_SET_BLANK:
                        dvb->videostate.videoBlank=(boolean) arg;
                        break;

                case VIDEO_GET_STATUS:
                        if (copy_to_user(parg, &dvb->videostate, 
					sizeof(struct videoStatus)))
                                return -EFAULT;
                        break;

                case VIDEO_GET_EVENT:
                        //FIXME: write firmware support for this
                        return -EOPNOTSUPP;

                case VIDEO_SET_DISPLAY_FORMAT:
                {
                        videoDisplayFormat_t format=(videoDisplayFormat_t) arg;
                        u16 val=0;
                        
                        switch(format) {
                        case VIDEO_PAN_SCAN:
                                val=VID_PAN_SCAN_PREF;
                                break;

                        case VIDEO_LETTER_BOX:
                                val=VID_VC_AND_PS_PREF;
                                break;

                        case VIDEO_CENTER_CUT_OUT:
                                val=VID_CENTRE_CUT_PREF;
                                break;

                        default:
                                return -EINVAL;
                        }

                        dvb->videostate.videoFormat=format;
                        return outcom(dvb, COMTYPE_ENCODER, SetPanScanType, 
                                      1, (u16) val);
                }

                case VIDEO_SET_FORMAT:
                        if (arg>1)
                                return -EINVAL;
                        dvb->display_ar=arg;
                        return outcom(dvb, COMTYPE_ENCODER, SetMonitorType, 
                                      1, (u16) arg);
                
                case VIDEO_STILLPICTURE:
		{ 
		        struct videoDisplayStillPicture pic;

                        if(copy_from_user(&pic, parg, 
                                          sizeof(struct videoDisplayStillPicture)))
                                return -EFAULT;

                        dvb_buf_flush(&dvb->avout);
                        AV_StartPlay(dvb, RP_VIDEO);
                        play_iframe(dvb, pic.iFrame, pic.size, 
                                    file->f_flags&O_NONBLOCK);
                        break;
		}

                case VIDEO_FAST_FORWARD:
                        if (dvb->videostate.streamSource!=VIDEO_SOURCE_MEMORY)
                                return -EPERM;
                        //note: arg is ignored by firmware
                        if (dvb->playing==RP_AV) 
                                outcom(dvb, COMTYPE_REC_PLAY, 
                                       __Scan_I, 2, AV_PES, 0);
                        else 
                                vidcom(dvb, 0x16, arg); 
                        dvb->trickmode=TRICK_FAST;
                        dvb->videostate.playState=VIDEO_PLAYING;
                        break;

                case VIDEO_SLOWMOTION:
                        if (dvb->videostate.streamSource!=VIDEO_SOURCE_MEMORY)
                                return -EPERM;
                        if (dvb->playing==RP_AV) {
                                outcom(dvb, COMTYPE_REC_PLAY, __Slow, 2, 0, 0);
                                vidcom(dvb, 0x22, arg);
                        } else {
                                vidcom(dvb, 0x0d, 0);
                                vidcom(dvb, 0x0e, 0);
                                vidcom(dvb, 0x22, arg);
                        }
                        dvb->trickmode=TRICK_SLOW;
                        dvb->videostate.playState=VIDEO_PLAYING;
                        break;

                case VIDEO_GET_CAPABILITIES:
                {
                        int cap=VIDEO_CAP_MPEG1|
                                VIDEO_CAP_MPEG2|
                                VIDEO_CAP_SYS|
                                VIDEO_CAP_PROG;
                        
                        if (copy_to_user(parg, &cap, 
					sizeof(cap)))
                                return -EFAULT;
                        break;
                }
                case VIDEO_CLEAR_BUFFER:
                        dvb_buf_flush(&dvb->avout);
                        reset_ipack(&dvb->ipack[1]);

                        if (dvb->playing==RP_AV) {
                                outcom(dvb, COMTYPE_REC_PLAY, 
                                       __Play, 2, AV_PES, 0);
                                if (dvb->trickmode==TRICK_FAST)
                                        outcom(dvb, COMTYPE_REC_PLAY, 
                                               __Scan_I, 2, AV_PES, 0);
                                if (dvb->trickmode==TRICK_SLOW) {
                                        outcom(dvb, COMTYPE_REC_PLAY, __Slow, 2, 0, 0);
                                        vidcom(dvb, 0x22, arg);
                                }
                                if (dvb->trickmode==TRICK_FREEZE)
                                        vidcom(dvb, 0x000e, 1);
                        }
                        break;
                case VIDEO_SET_STREAMTYPE:
                        
                        break;
                default:
                        return -ENOIOCTLCMD;
		}
                return 0;

	case DVB_DEVICE_AUDIO_0:
                if (((file->f_flags&O_ACCMODE)==O_RDONLY) &&
                    (cmd!=AUDIO_GET_STATUS))
                        return -EPERM;

                switch (cmd) {

                case AUDIO_STOP:
                        if (dvb->audiostate.streamSource==AUDIO_SOURCE_MEMORY)
                                AV_Stop(dvb, RP_AUDIO);
                        else
                                audcom(dvb, 1);
                        dvb->audiostate.playState=AUDIO_STOPPED;
                        break;

                case AUDIO_PLAY:
                        if (dvb->audiostate.streamSource==AUDIO_SOURCE_MEMORY)
                                AV_StartPlay(dvb, RP_AUDIO);
                        audcom(dvb, 2);
                        dvb->audiostate.playState=AUDIO_PLAYING;
		        break;

                case AUDIO_PAUSE:
			audcom(dvb, 1);
		        dvb->audiostate.playState=AUDIO_PAUSED;
                        break;

                case AUDIO_CONTINUE:
		        if (dvb->audiostate.playState==AUDIO_PAUSED) {
			        dvb->audiostate.playState=AUDIO_PLAYING;
				audcom(dvb, 0x12);
			} 
		        break;

                case AUDIO_SELECT_SOURCE:
                        dvb->audiostate.streamSource=(audioStreamSource_t) arg;
		        break;

                case AUDIO_SET_MUTE:
                {
			audcom(dvb, arg ? 1 : 2);
                        dvb->audiostate.muteState=(boolean) arg;
                        break;
		}

                case AUDIO_SET_AV_SYNC:
                        dvb->audiostate.AVSyncState=(boolean) arg;
			audcom(dvb, arg ? 0x0f : 0x0e);
                        break;

                case AUDIO_SET_BYPASS_MODE:
		        return -EINVAL;

                case AUDIO_CHANNEL_SELECT:
		        dvb->audiostate.channelSelect=(audioChannelSelect_t) arg;
                        
                        switch(dvb->audiostate.channelSelect) {
                        case AUDIO_STEREO:
  			        audcom(dvb, 0x80);
                                break;

                        case AUDIO_MONO_LEFT:
  			        audcom(dvb, 0x100);
                                break;

                        case AUDIO_MONO_RIGHT:
  			        audcom(dvb, 0x200);
                                break;

                        default:
                                return -EINVAL;
			}
			return 0;

                case AUDIO_GET_STATUS:
                        if(copy_to_user(parg, &dvb->audiostate, 
					sizeof(struct audioStatus)))
                                return -EFAULT;
                        break;

                case AUDIO_GET_CAPABILITIES:
                {
                        int cap=AUDIO_CAP_LPCM|
                                AUDIO_CAP_MP1|
                                AUDIO_CAP_MP2;
                        
                        if (copy_to_user(parg, &cap, 
					sizeof(cap)))
                                return -EFAULT;
                        break;
                }
                case AUDIO_CLEAR_BUFFER:
                        dvb_buf_flush(&dvb->aout);
                        reset_ipack(&dvb->ipack[0]);
                        if (dvb->playing==RP_AV)
                                outcom(dvb, COMTYPE_REC_PLAY, 
                                       __Play, 2, AV_PES, 0);
                        break;
                case AUDIO_SET_ID:
                  
                        break;
                case AUDIO_SET_MIXER:
                {
		        struct audioMixer amix;

                        if(copy_from_user(&amix, parg, 
                                          sizeof(struct audioMixer)))
                                return -EFAULT;
                        SetVolume(dvb, amix.volume_left, amix.volume_right);
                }
                        break;
                case AUDIO_SET_STREAMTYPE:
                        
                        break;
                default:
                        return -ENOIOCTLCMD;
                }
                return 0;
       
	case DVB_DEVICE_QAMFE_0:
                switch (cmd) {
                case OST_SELFTEST:
                        break;

                case OST_SET_POWER_STATE:
                        switch (arg) {
                        case OST_POWER_ON:
                                dvb->powerstate=arg;
                                return 0;
                        case OST_POWER_SUSPEND:
                        case OST_POWER_STANDBY:
                        case OST_POWER_OFF:
                                dvb->powerstate=OST_POWER_OFF;
                                return 0;
                        default:
                                return -EINVAL;
                        }
                        break;
                        
                case OST_GET_POWER_STATE:
                        if(copy_to_user(parg, &dvb->powerstate, sizeof(u32)))
                                 return -EFAULT;
                        break;

                case FE_READ_STATUS:
                {
                        feStatus stat;
                        
                        GetFront(dvb);
                        stat=0;
                        if (dvb->front.power)
                                stat|=FE_HAS_POWER;
                        if ((dvb->front.sync&0x1f)==0x1f)
                                stat|=FE_HAS_SIGNAL;
                        if ((dvb->front.sync&1))
                                stat|=FE_HAS_LOCK;
                        if (dvb->front.inv)
                                stat|=QPSK_SPECTRUM_INV;
                        if(copy_to_user(parg, &stat, sizeof(stat)))
                                return -EFAULT;
                        break;
                }

                case FE_READ_BER:
                {
                        uint32_t ber;
                        
                        GetFront(dvb);
                        if (!dvb->front.power)
                                return -ENOSIGNAL;
                        ber=dvb->front.vber*10;
                        if(copy_to_user(parg, &ber, sizeof(ber)))
                                return -EFAULT;
                        break;
                }
                
                case FE_READ_SIGNAL_STRENGTH:
                {
                        int32_t signal;
                        
                        GetFront(dvb);
                        if (!dvb->front.power)
                                return -ENOSIGNAL;
                        signal=0xff-(dvb->front.agc>>8); //FIXME: scale value
                        if(copy_to_user(parg, &signal, sizeof(signal)))
                                return -EFAULT;
                        break;
                }

                case FE_READ_SNR:
                {
                        int32_t snr;
                        
                        GetFront(dvb);
                        if (!dvb->front.power)
                                return -ENOSIGNAL;
                        //linear approximation, FIXME: make nest2snr table
                        snr=20000000+(10-(dvb->front.nest>>8))*20000000/160;
                        //snr=nest2snr[dvb->front.nest];
                        if(copy_to_user(parg, &snr, sizeof(snr)))
                                return -EFAULT;
                        break;
                }
                
                case FE_READ_UNCORRECTED_BLOCKS: 
                  //FIXME: VES1893 only supports counter up to 0x7f
                  //       how should we handle this??
                        return -ENOSYS; 
                
                case FE_GET_NEXT_FREQUENCY:
                {
                        uint32_t freq;
                                    
                        if (copy_from_user(&freq, parg, sizeof(freq)))
                                return -EFAULT;
                
                        freq+=1000000;
                        
                        if (copy_to_user(parg, &freq, sizeof(freq)))
                            return -EFAULT;
                        break;
                }
               
                case FE_GET_NEXT_SYMBOL_RATE:
                {
                        uint32_t rate;
                        
                        if(copy_from_user(&rate, parg, sizeof(rate)))
                                return -EFAULT;
                        // FIXME: how does one really calculate this?
                        if (rate<5000000)
                                rate+=500000;
                        else if(rate<10000000)
                                rate+=1000000;
                        else if(rate<30000000)
                                rate+=2000000;
                        else 
                                return -EINVAL;
                        if(copy_to_user(parg, &rate, sizeof(rate)))
                                return -EFAULT;
                        break;
                }

                case QAM_GET_STATUS:
                {
                        struct qamParameters para;

                        para.Frequency=dvb->front.freq/1000;
                        para.SymbolRate=dvb->front.srate;
                        para.FEC_inner=fectab2[dvb->front.fec];
                        para.QAM=dvb->front.qam;

                        if(copy_to_user(parg, &para, sizeof(para)))
                                return -EFAULT;
                        break;
                }
                case QAM_TUNE:
                {
                        struct qamParameters para;
                        
                        if(copy_from_user(&para, parg, sizeof(para)))
                                return -EFAULT;
                        if ((file->f_flags&O_ACCMODE)==O_RDONLY) 
                                return -EPERM;

                        return qam_tune(dvb, &para);
                }
                
                case QAM_GET_EVENT:
                { 
                        int ret;

                        if (dvb->qam.overflow) {
                                dvb->qam.overflow=0;
                                return -EBUFFEROVERFLOW;
                        }
                        if (dvb->qam.eventw==dvb->qam.eventr) {
                                if (file->f_flags&O_NONBLOCK) 
                                        return -EWOULDBLOCK;
                                else {
                                  ret=wait_event_interruptible(dvb->qam.eventq,
                                                               dvb->qam.eventw!=
                                                               dvb->qam.eventr);
                                  if (ret<0)
                                          return ret;
                                }
                        }
                        spin_lock(&dvb->qam.eventlock);

                        if(copy_to_user(parg, &dvb->qam.events[dvb->qam.eventr], 
                                        sizeof(struct qamEvent))) {
                                spin_unlock(&dvb->qam.eventlock);
                                return -EFAULT;
                        }
                        dvb->qam.eventr=(dvb->qam.eventr+1)%QAM_MAX_EVENT;
                        spin_unlock(&dvb->qam.eventlock);
                        break;
                }
                
                case QAM_FE_INFO:
                {        
                        struct qamFrontendInfo feinfo;
                        
                        feinfo.minFrequency=40000; 
                        feinfo.maxFrequency=870000;
                        feinfo.minSymbolRate=500000;
                        feinfo.maxSymbolRate=30000000;
                        feinfo.hwType=0;    //??
                        feinfo.hwVersion=0;

                        if(copy_to_user(parg, &feinfo, sizeof(feinfo)))
                                return -EFAULT;
                        break;
                }

                case QAM_WRITE_REGISTER:
                {
                        struct qamRegister qamr;

                        if(copy_from_user(&qamr, parg, sizeof(qamr)))
                                return -EFAULT;
                        switch(qamr.chipId) {
                        case 0x20:  
                                demod_writereg(dvb, qamr.address, qamr.value);
                                break;
                        default:
                                return -EINVAL;
                        }
                        break;
                }
                

                case QAM_READ_REGISTER:
                {
                        struct qamRegister qamr;

                        if(copy_from_user(&qamr, parg, sizeof(qamr)))
                                return -EFAULT;
                        switch(qamr.chipId) {
                        case 0x20:  
                                qamr.value=demod_readreg(dvb, qamr.address);
                                break;
                        default:
                                return -EINVAL;
                        }
                        if (copy_to_user(parg, &qamr, sizeof(qamr)))
                                return -EFAULT;
                        break;
                }

                default:
                        return -ENOIOCTLCMD;
                
                }
                return 0;

	case DVB_DEVICE_QPSKFE_0:

                switch (cmd) {
                case OST_SELFTEST:
                        break;

                case OST_SET_POWER_STATE:
                        switch (arg) {
                        case OST_POWER_ON:
                                dvb->powerstate=arg;
                                SetVoltage(dvb,1,0);
                                return 0;
                        case OST_POWER_SUSPEND:
                        case OST_POWER_STANDBY:
                        case OST_POWER_OFF:
                                dvb->powerstate=OST_POWER_OFF;
                                SetVoltage(dvb,0,0);
                                return 0;
                        default:
                                return -EINVAL;
                        }
                        break;
                        
                case OST_GET_POWER_STATE:
                        if(copy_to_user(parg, &dvb->powerstate, sizeof(u32)))
                                 return -EFAULT;
                        break;

                case FE_READ_STATUS:
                {
                        feStatus stat;
                        
                        GetFront(dvb);
                        stat=0;
                        if (dvb->front.power)
                                stat|=FE_HAS_POWER;
                        if ((dvb->front.sync&0x1f)==0x1f)
                                stat|=FE_HAS_SIGNAL;
                        if ((dvb->front.sync&1))
                                stat|=FE_HAS_LOCK;
                        if (dvb->front.inv)
                                stat|=QPSK_SPECTRUM_INV;
                        if(copy_to_user(parg, &stat, sizeof(stat)))
                                return -EFAULT;
                        break;
                }

                case FE_READ_BER:
                {
                        uint32_t ber;
                        
                        GetFront(dvb);
                        if (!dvb->front.power)
                                return -ENOSIGNAL;
                        ber=dvb->front.vber*10;
                        if(copy_to_user(parg, &ber, sizeof(ber)))
                                return -EFAULT;
                        break;
                }
                
                case FE_READ_SIGNAL_STRENGTH:
                {
                        int32_t signal;
                        
                        GetFront(dvb);
                        if (!dvb->front.power)
                                return -ENOSIGNAL;
                        signal=0xff-(dvb->front.agc>>8); //FIXME: scale value
                        if(copy_to_user(parg, &signal, sizeof(signal)))
                                return -EFAULT;
                        break;
                }

                case FE_READ_SNR:
                {
                        int32_t snr;
                        
                        GetFront(dvb);
                        if (!dvb->front.power)
                                return -ENOSIGNAL;
                        //linear approximation, FIXME: make nest2snr table
                        snr=20000000+(10-(dvb->front.nest>>8))*20000000/160;
                        //snr=nest2snr[dvb->front.nest];
                        if(copy_to_user(parg, &snr, sizeof(snr)))
                                return -EFAULT;
                        break;
                }
                
                case FE_READ_UNCORRECTED_BLOCKS: 
                  //FIXME: VES1893 only supports counter up to 0x7f
                  //       how should we handle this??
                        return -ENOSYS; 
                
                case FE_GET_NEXT_FREQUENCY:
                {
                        uint32_t freq;
                                    
                        if (copy_from_user(&freq, parg, sizeof(freq)))
                                return -EFAULT;
                
                        // FIXME: how does one calculate this?
                        freq+=1000; //FIXME: KHz like in QPSK_TUNE??
                        
                        if (copy_to_user(parg, &freq, sizeof(freq)))
                            return -EFAULT;
                        break;
                }
               
                case FE_GET_NEXT_SYMBOL_RATE:
                {
                        uint32_t rate;
                        
                        if(copy_from_user(&rate, parg, sizeof(rate)))
                                return -EFAULT;
                        // FIXME: how does one really calculate this?
                        if (rate<5000000)
                                rate+=500000;
                        else if(rate<10000000)
                                rate+=1000000;
                        else if(rate<30000000)
                                rate+=2000000;
                        else 
                                return -EINVAL;
                        if(copy_to_user(parg, &rate, sizeof(rate)))
                                return -EFAULT;
                        break;
                }

                case QPSK_GET_STATUS:
                {
                        struct qpskParameters para;

                        para.iFrequency=dvb->front.freq/1000;
                        para.SymbolRate=dvb->front.srate;
                        para.FEC_inner=fectab2[dvb->front.fec];

                        if(copy_to_user(parg, &para, sizeof(para)))
                                return -EFAULT;
                }

                case QPSK_TUNE:
                {
                        struct qpskParameters para;
                        
                        if(copy_from_user(&para, parg, sizeof(para)))
                                return -EFAULT;
                        if ((file->f_flags&O_ACCMODE)==O_RDONLY) 
                                return -EPERM;

                        return qpsk_tune(dvb, &para);
                }
                
                case QPSK_GET_EVENT:
                { 
                        int ret;

                        if (dvb->qpsk.overflow) {
                                dvb->qpsk.overflow=0;
                                return -EBUFFEROVERFLOW;
                        }
                        if (dvb->qpsk.eventw==dvb->qpsk.eventr) {
                                if (file->f_flags&O_NONBLOCK) 
                                        return -EWOULDBLOCK;
                                else {
                                        ret=wait_event_interruptible(
                                                dvb->qpsk.eventq,
                                                dvb->qpsk.eventw!=
                                                dvb->qpsk.eventr);
                                        if (ret<0)
                                                return ret;
                                }
                        }
                        spin_lock(&dvb->qpsk.eventlock);
                        if (copy_to_user(parg, 
                                         &dvb->qpsk.events[dvb->qpsk.eventr], 
                                         sizeof(struct qpskEvent))) {
                                spin_unlock(&dvb->qpsk.eventlock);
                                return -EFAULT;
                        }
                        dvb->qpsk.eventr=(dvb->qpsk.eventr+1)%QPSK_MAX_EVENT;
                        spin_unlock(&dvb->qpsk.eventlock);
                        break;
                }
                
                case QPSK_FE_INFO:
                {        
                        struct qpskFrontendInfo feinfo;
                        
                        feinfo.minFrequency=500;  //KHz?
                        feinfo.maxFrequency=2700000;
                        feinfo.minSymbolRate=500000;
                        feinfo.maxSymbolRate=30000000;
                        feinfo.hwType=0;    //??
                        feinfo.hwVersion=0;

                        if(copy_to_user(parg, &feinfo, sizeof(feinfo)))
                                return -EFAULT;
                        break;
                }

                case QPSK_WRITE_REGISTER:
                {
                        struct qpskRegister qpskr;

                        if(copy_from_user(&qpskr, parg, sizeof(qpskr)))
                                return -EFAULT;
                        switch(qpskr.chipId) {
                        case 0x20:  
                                demod_writereg(dvb, qpskr.address, qpskr.value);
                                break;
                        default:
                                return -EINVAL;
                        }
                        break;
                }
                

                case QPSK_READ_REGISTER:
                {
                        struct qpskRegister qpskr;

                        if(copy_from_user(&qpskr, parg, sizeof(qpskr)))
                                return -EFAULT;
                        switch(qpskr.chipId) {
                        case 0x20:  
                                qpskr.value=demod_readreg(dvb, qpskr.address);
                                break;
                        default:
                                return -EINVAL;
                        }
                        if (copy_to_user(parg, &qpskr, sizeof(qpskr)))
                                return -EFAULT;
                        break;
                }

                default:
                        return -ENOIOCTLCMD;
                
                }
                return 0;
        // NOTE: right now the DiSEqC support in the firmware is set up
        //       as busy waiting so that atomicity is guaranteed automatically

	case DVB_DEVICE_SEC_0:
                if (file->f_flags&O_NONBLOCK) 
                        return -EWOULDBLOCK;
                switch (cmd) {
                case SEC_GET_STATUS: 
                {        
                        struct secStatus status;
                        
                        status.busMode=SEC_BUS_IDLE;
                        
                        if (dvb->secbusy) 
                                status.busMode=SEC_BUS_BUSY;
                        if (!dvb->front.power) 
                                status.busMode=SEC_BUS_OFF;
                        
                        status.selVolt=(dvb->front.volt ?
                                        SEC_VOLTAGE_18 :
                                        SEC_VOLTAGE_13);

                        status.contTone=(dvb->front.ttk ?
                                         SEC_TONE_ON :
                                         SEC_TONE_OFF);
                        if(copy_to_user(parg, &status, sizeof(status)))
                                return -EFAULT;
                }
                        break;
                case SEC_RESET_OVERLOAD:
                        if ((file->f_flags&O_ACCMODE)==O_RDONLY) 
                                return -EPERM;
                        // overload control not supported by Siemens DVB card
                        break;

                case SEC_SEND_SEQUENCE:
                {
                        struct secCmdSequence seq;

                        if(copy_from_user(&seq, parg, sizeof(seq)))
                                return -EFAULT;

                        if ((file->f_flags&O_ACCMODE)==O_RDONLY) 
                                return -EPERM;

                        mon_stop(dvb);
                        return secSendSequence(dvb, &seq);
                }
                case SEC_SET_TONE:
                {
                        secToneMode mode = (secToneMode) arg;

                        if ((file->f_flags&O_ACCMODE)==O_RDONLY) 
                                return -EPERM;
                        
                        mon_stop(dvb);
                        return secSetTone(dvb, mode);
                }
                case SEC_SET_VOLTAGE:
                {
                        secVoltage val = (secVoltage) arg;

                        if ((file->f_flags&O_ACCMODE)==O_RDONLY) 
                                return -EPERM;
                        
                        mon_stop(dvb);
                        return secSetVoltage(dvb, val);
                }
                default:
                        return -EINVAL;
                }
		return 0;
                
        case DVB_DEVICE_DEMUX_0:
                return DmxDevIoctl(&dvb->dmxdev, file, cmd, arg);

        case DVB_DEVICE_CA_0:
                switch (cmd) {
                case CA_RESET:
                  
                      break;

                case CA_GET_CAP:
                {
                        ca_cap_t cap;

                        cap.slot_num=2;
                        cap.slot_type=CA_CI;
                        cap.descr_num=32;
                        cap.descr_type=CA_ECD;
                        if(copy_to_user(parg, &cap, sizeof(cap)))
                                return -EFAULT;
                }
                        break;
                case CA_GET_SLOT_INFO:
                {
                        ca_slot_info_t info;

                        if(copy_from_user(&info, parg, sizeof(info)))
                                return -EFAULT;
                        if (info.num>1)
                                return -EINVAL;
                        info.type=CA_CI;
                        info.flags=0;
                        if (dvb->ci_slot[info.type]&CI_MODULE_PRESENT)
                                info.flags|=CA_CI_MODULE_PRESENT;
                        if (dvb->ci_slot[info.type]&CI_MODULE_OK)
                                info.flags|=CA_CI_MODULE_READY;
                                                
                        if(copy_to_user(parg, &info, sizeof(info)))
                                return -EFAULT;

                }
                        break;
                case CA_GET_MSG:

                        break;
                case CA_SEND_MSG:

                        break;

                case CA_SET_DESCR:
                {
                        ca_descr_t descr;
                        
                        if(copy_from_user(&descr, parg, sizeof(descr)))
                                return -EFAULT;
                        if (descr.index>=12)
                                return -EINVAL;
                        if (descr.parity>1)
                                return -EINVAL;
                        outcom(dvb, COMTYPE_PIDFILTER, SetDescr, 5, 
                               (descr.index<<8)|descr.parity,
                               (descr.cw[0]<<8)|descr.cw[1],
                               (descr.cw[2]<<8)|descr.cw[3],
                               (descr.cw[4]<<8)|descr.cw[5],
                               (descr.cw[6]<<8)|descr.cw[7]);
                }
                        break;
                default:
                        return -EINVAL;
                }
                break;
        case DVB_DEVICE_OSD_0:
        {
#ifdef USE_OSD
                switch (cmd) {
                case OSD_SEND_CMD:
                {
                        osd_cmd_t doc;

                        if(copy_from_user(&doc, parg, sizeof(osd_cmd_t)))
                                return -EFAULT;
                        return OSD_DrawCommand(dvb, &doc);
                }
                default:
                        return -EINVAL;
                }
#endif
                break;
        }
	case DVB_DEVICE_NET_0:
        {
                if (((file->f_flags&O_ACCMODE)==O_RDONLY))
                        return -EPERM;
                
                switch (cmd) {
                case NET_ADD_IF:
                {
                        struct dvb_net_if dvbnetif;
                        int result;

                        if(copy_from_user(&dvbnetif, parg, sizeof(dvbnetif)))
                                return -EFAULT;
                        result=dvb_net_add_if(&dvb->dvb_net, dvbnetif.pid);
                        if (result<0)
                                return result;
                        dvbnetif.if_num=result;
                        if(copy_to_user(parg, &dvbnetif, sizeof(dvbnetif)))
                                return -EFAULT;
                        break;
                }
                case NET_REMOVE_IF:
                        return dvb_net_remove_if(&dvb->dvb_net, (int) arg);
                default:
                        return -EINVAL;
                }
                break;
        }
        default:
                return -EOPNOTSUPP;
        }
        return 0;
}

static unsigned int 
dvbdev_poll(struct dvb_device *dvbdev, int num, 
            struct file *file, poll_table * wait)
{
        struct dvb_struct *dvb=(struct dvb_struct *) dvbdev->priv;
        int type=num2type(dvb, num);

	switch (type) {
	case DVB_DEVICE_QPSKFE_0:
                if (dvb->qpsk.eventw!=dvb->qpsk.eventr)
                        return (POLLIN | POLLRDNORM | POLLPRI);

                poll_wait(file, &dvb->qpsk.eventq, wait);
                
                if (dvb->qpsk.eventw!=dvb->qpsk.eventr)
                        return (POLLIN | POLLRDNORM | POLLPRI);
                return 0;
	case DVB_DEVICE_QAMFE_0:
                if (dvb->qam.eventw!=dvb->qam.eventr)
                        return (POLLIN | POLLRDNORM | POLLPRI);

                poll_wait(file, &dvb->qam.eventq, wait);
                
                if (dvb->qam.eventw!=dvb->qam.eventr)
                        return (POLLIN | POLLRDNORM | POLLPRI);
                return 0;

        case DVB_DEVICE_DEMUX_0:
                return DmxDevPoll(&dvb->dmxdev, file, wait);

	case DVB_DEVICE_DVR_0:
                return DmxDevDVRPoll(&dvb->dmxdev, file, wait);

        case DVB_DEVICE_VIDEO_0:
                return dvb_poll(dvb, file, wait);
                
        case DVB_DEVICE_AUDIO_0:
                return dvb_apoll(dvb, file, wait);

        case DVB_DEVICE_CA_0:
                break;

	default:
	        return -EOPNOTSUPP;
        }

        return 0;
}


static int 
dvbdev_device_type(struct dvb_device *dvbdev, unsigned int num)
{
        struct dvb_struct *dvb=(struct dvb_struct *) dvbdev->priv;

        return num2type(dvb, num);
}

/******************************************************************************
 * driver registration 
 ******************************************************************************/

static int dvb_register(struct dvb_struct *dvb)
{
        int ret, i;
        struct dvb_device *dvbd=&dvb->dvb_dev;
        dmx_frontend_t *dvbfront=&dvb->hw_frontend;
        dvb_demux_t *dvbdemux=&dvb->demux;

        if (dvb->registered)
                return -1;
        dvb->registered=1;

        init_waitqueue_head(&dvb->qpsk.eventq);
        spin_lock_init (&dvb->qpsk.eventlock);
        dvb->qpsk.eventw=dvb->qpsk.eventr=0;
        dvb->qpsk.overflow=0;
        init_waitqueue_head(&dvb->qam.eventq);
        spin_lock_init (&dvb->qam.eventlock);
        dvb->qam.eventw=dvb->qam.eventr=0;
        dvb->qam.overflow=0;
        dvb->secbusy=0;

	dvb->audiostate.AVSyncState=0;
	dvb->audiostate.muteState=0;
	dvb->audiostate.playState=AUDIO_STOPPED;
	dvb->audiostate.streamSource=AUDIO_SOURCE_DEMUX;
	dvb->audiostate.channelSelect=AUDIO_STEREO;
	dvb->audiostate.bypassMode=0;

	dvb->videostate.videoBlank=0;
	dvb->videostate.playState=VIDEO_STOPPED;
	dvb->videostate.streamSource=VIDEO_SOURCE_DEMUX;
	dvb->videostate.videoFormat=VIDEO_FORMAT_4_3;
	dvb->videostate.displayFormat=VIDEO_CENTER_CUT_OUT;
        dvb->display_ar=VIDEO_FORMAT_4_3;

        memcpy(dvb->demux_id, "demux0_0", 9);
        dvb->demux_id[7]=dvb->num+0x30;
        dvbdemux->priv=(void *) dvb;

        if (dvb->card_type==DVB_CARD_TT_SIEMENS) {
                for (i=0; i<32; i++)
                        dvb->handle2filter[i]=NULL;

                dvbdemux->filternum=32;
                dvbdemux->feednum=32;
                dvbdemux->start_feed=dvb_start_feed;
                dvbdemux->stop_feed=dvb_stop_feed;
                dvbdemux->write_to_decoder=dvb_write_to_decoder;
                
                dvbdemux->dmx.vendor="TI";
                dvbdemux->dmx.model="AV7110";
                dvbdemux->dmx.id=dvb->demux_id;
                dvbdemux->dmx.capabilities=(DMX_TS_FILTERING|
                                             DMX_SECTION_FILTERING|
                                             DMX_MEMORY_BASED_FILTERING);

                DvbDmxInit(&dvb->demux);


                dvbfront->id="hw_frontend";
                dvbfront->vendor="VLSI";

                switch (dvb->dvbtype) {
                default:
                case DVB_S:
                        dvbfront->model="DVBS";
                        dvbfront->source=DMX_SATELLITE_FE;
                        break;
                case DVB_C:
                        dvbfront->model="1820";
                        dvbfront->source=DMX_CABLE_FE;
                        break;
                }
                dvb->dmxdev.filternum=32;
                dvb->dmxdev.demux=&dvbdemux->dmx;
                dvb->dmxdev.capabilities=0;

                DmxDevInit(&dvb->dmxdev);
        }

        if (dvb->card_type==DVB_CARD_TT_BUDGET) {
                dvbdemux->filternum=256;
                dvbdemux->feednum=256;
                dvbdemux->start_feed=dvb_start_feed;
                dvbdemux->stop_feed=dvb_stop_feed;
                dvbdemux->write_to_decoder=0;

                dvbdemux->dmx.vendor="CIM";
                dvbdemux->dmx.model="sw";
                dvbdemux->dmx.id=dvb->demux_id;
                dvbdemux->dmx.capabilities=(DMX_TS_FILTERING|
                                             DMX_SECTION_FILTERING|
                                             DMX_MEMORY_BASED_FILTERING);

                DvbDmxInit(&dvb->demux);

                dvbfront->id="hw_frontend";
                dvbfront->vendor="VLSI";
                switch (dvb->dvbtype) {
                default:
                case DVB_S:
                        dvbfront->model="DVBS";
                        dvbfront->source=DMX_SATELLITE_FE;
                        break;
                case DVB_C:
                        dvbfront->model="1820";
                        dvbfront->source=DMX_CABLE_FE;
                        break;
                }
                
                dvb->dmxdev.filternum=256;
                dvb->dmxdev.demux=&dvbdemux->dmx;
                dvb->dmxdev.capabilities=0;

                DmxDevInit(&dvb->dmxdev);
        }

        ret=dvbdemux->dmx.add_frontend(&dvbdemux->dmx, 
                                        &dvb->hw_frontend);
        if (ret<0)
                return ret;
        
        dvb->mem_frontend.id="mem_frontend";
        dvb->mem_frontend.vendor="memory";
        dvb->mem_frontend.model="sw";
        dvb->mem_frontend.source=DMX_MEMORY_FE;
        ret=dvbdemux->dmx.add_frontend(&dvbdemux->dmx, 
                                        &dvb->mem_frontend);
        if (ret<0)
                return ret;
        
        ret=dvbdemux->dmx.connect_frontend(&dvbdemux->dmx, 
                                            &dvb->hw_frontend);
        if (ret<0)
                return ret;

        // init and register dvb device structure
        dvbd->priv=(void *) dvb;
        dvbd->open=dvbdev_open;
        dvbd->close=dvbdev_close;
        dvbd->read=dvbdev_read;
        dvbd->write=dvbdev_write;
        dvbd->ioctl=dvbdev_ioctl;
        dvbd->poll=dvbdev_poll;
        dvbd->device_type=dvbdev_device_type;
        
        for (i=0; i<DVB_DEVS_MAX; i++) 
                dvb->users[i]=dvb->writers[i]=0;

        dvb->dvb_devs=0;
        if (dvb->card_type==DVB_CARD_TT_SIEMENS) {
                if (dvb->dvbtype == DVB_S)
                        dvb->dvb_devs=&dvbs_devs;
                if (dvb->dvbtype == DVB_C)
                        dvb->dvb_devs=&dvbc_devs;
        }
        if (dvb->card_type==DVB_CARD_TT_BUDGET) {
                if (dvb->dvbtype == DVB_S)
                        dvb->dvb_devs=&ttbs_devs;
                if (dvb->dvbtype == DVB_C)
                        dvb->dvb_devs=&ttbc_devs;
        }
        
        dvb->dvb_net.card_num=dvb->num;
        dvb_net_init(&dvb->dvb_net, &dvbdemux->dmx);
        return dvb_register_device(dvbd);
}


static void dvb_unregister(struct dvb_struct *dvb)
{
        dvb_demux_t *dvbdemux=&dvb->demux;

        if (!dvb->registered)
                return;
        dvb_net_release(&dvb->dvb_net);
        dvbdemux->dmx.close(&dvbdemux->dmx);
        dvbdemux->dmx.remove_frontend(&dvbdemux->dmx, &dvb->hw_frontend);
        dvbdemux->dmx.remove_frontend(&dvbdemux->dmx, &dvb->mem_frontend);

        DmxDevRelease(&dvb->dmxdev);
        DvbDmxRelease(&dvb->demux);
        dvb_unregister_device(&dvb->dvb_dev);
}

/****************************************************************************
 * INITIALIZATION
 ****************************************************************************/

static int __init
get_card_type(struct pci_dev *dev)
{
        //FIXME: change back to -1 when we are sure to know all subvendor device ids
        int type=DVB_CARD_TT_SIEMENS; 
        u16 subvendor, subdevice;

        pci_read_config_word(dev, PCI_SUBSYSTEM_VENDOR_ID, &subvendor);
        pci_read_config_word(dev, PCI_SUBSYSTEM_ID, &subdevice);
        
        if (subvendor==0x13c2) {
                switch (subdevice) {
                case 0x1003: /* TT-Budget and WinTV-NOVA, both tuner types, DVBS */
                case 0x1004: /* WinTV-NOVA, DVBC */
                        type=DVB_CARD_TT_BUDGET;
                        break;                  
                case 0x0000: /* Siemens cable and sat card (maybe older version?) */
                case 0x0001: /* Technotrend, Hauppauge (maybe retail version ?)   */
                case 0x0003: /* Hauppauge, new design and new tuner */
                case 0x0004: /* new type but with old tuner ? */
                        type=DVB_CARD_TT_SIEMENS;
                        break;                  
                }
        }
        return type;
}


__init static int 
init_card(struct dvb_struct *dvb, int num, 
          struct i2c_adapter *adap, int card_type)
{
        int i;

        memset(dvb, 0, sizeof(struct dvb_struct));
        dvb->card_type=card_type;

        spin_lock_init (&dvb->debilock);
        sema_init(&dvb->dcomlock, 1);
        init_waitqueue_head(&dvb->debiq);
        dvb->debilock=SPIN_LOCK_UNLOCKED;

        dvb->num=num;
        dvb->i2cbus=adap;
        dvb->saa=(struct saa7146 *) (adap->data);

        dvb->dvbtype = DVB_S;
        dvb->adac_type = adac;

        dvb->osdwin=1;
	init_waitqueue_head(&dvb->mon_wait);
        sema_init(&dvb->mon_sem, 1);
	init_timer(&dvb->mon_timer);
        dvb->mon_timer.function=mon_timer;
        dvb->mon_timer.data=(unsigned long) dvb;
	kernel_thread(mon_thread, (void *)dvb, 0);

        dvb->debitype=-1;

        if(vid_register(dvb) < 0)
                return -1;

        dvb->vidmode=VIDEO_MODE_PAL;

        init_ipack(&dvb->ipack[0], IPACKS, play_audio_cb);
        dvb->ipack[0].data=(void *) dvb;
        init_ipack(&dvb->ipack[1], IPACKS, play_video_cb);
        dvb->ipack[1].data=(void *) dvb;
        

        /* allocate and init buffers */

        dvb->iobuf=vmalloc(AVOUTLEN+AOUTLEN+BMPLEN+4*IPACKS);
        if (!dvb->iobuf) 
                return -1;

        dvb_buf_init(&dvb->avout, dvb->iobuf, AVOUTLEN);
        dvb_buf_init(&dvb->aout, dvb->iobuf+AVOUTLEN, 
                     AOUTLEN);

        /* init BMP buffer */
        dvb->bmpbuf=dvb->iobuf+AVOUTLEN+AOUTLEN;
        init_waitqueue_head(&dvb->bmpq);
        
        dvb->kbuf[0]=(u8 *)(dvb->iobuf+AVOUTLEN+AOUTLEN+BMPLEN);
        dvb->kbuf[1]=dvb->kbuf[0]+2*IPACKS;


        /* handle different card types */

        dvb->tuner_type = -1;

        if (dvb->card_type==DVB_CARD_TT_SIEMENS)
                bootarm(dvb);

        if (dvb->card_type==DVB_CARD_TT_BUDGET) {
                saa7146_write(dvb->saa->mem, DD1_INIT, 0x02000600);
                saa7146_write(dvb->saa->mem, MC2, 
                              (MASK_09 | MASK_25 | MASK_10 | MASK_26));
                setgpio(dvb, 2, GPIO_OUTHI);
        }

        for (i=0; i<adap->client_count; i++) {
                if (!adap->clients[i])
                        continue;
                dprintk("adapter id = %02x  addr = %02x\n",
                        adap->clients[i]->driver->id,
                        adap->clients[i]->addr*2);
                
                switch (adap->clients[i]->driver->id) {
                case I2C_DRIVERID_TUNER:
                        dvb->tuner = adap->clients[i];
                        break;;
                case I2C_DRIVERID_VES1893:
                        dvb->dvbtype = DVB_S;
                        dvb->board_rev= 0x0103;
                        dvb->demod_type = DVB_DEMOD_VES1893;
                        dvb->tuner_type = TUNER_SP5659;
                        dvb->demod = adap->clients[i];
                        break;
                case I2C_DRIVERID_VES1820:
                        dvb->dvbtype = DVB_C;
                        dvb->demod_type = DVB_DEMOD_VES1820;
                        dvb->tuner_type = TUNER_DVBC;
                        dvb->demod = adap->clients[i];
                        break;
                case I2C_DRIVERID_SAA7113:
                        printk("dvb%d: SAA7113 detected\n", dvb->num);
                        break;
                case I2C_DRIVERID_STV0299:
                        dvb->dvbtype = DVB_S;
                        dvb->board_rev= 0x0201;
                        dvb->demod_type = DVB_DEMOD_STV0299;
                        dvb->tuner_type = TUNER_TSA5059;
                        dvb->demod = adap->clients[i];
                        break;
                case I2C_DRIVERID_TDA8083:
                        dvb->dvbtype = DVB_S;
                        dvb->board_rev= 0x0106;
                        dvb->demod_type = DVB_DEMOD_TDA8083;
                        dvb->tuner_type = TUNER_TSA5522;
                        dvb->adac_type = DVB_ADAC_CRYSTAL;
                        dvb->demod = adap->clients[i];
                        break;
                }
        }
        dvb_register(dvb);

        if (dvb->tuner && dvb->demod) 
                InitFront(dvb);

        if (dvb->card_type==DVB_CARD_TT_BUDGET) 
                return 0;

        if ((dvb->dvbtype == DVB_C) &&
            (i2c_writereg(dvb, 0x80, 0x0, 0x80)==1)) {
                i2c_writereg(dvb, 0x80, 0x0, 0);
                printk("dvb%d: DVB-C analog module detected, "
                       "initializing MSP3400\n", dvb->num);
                ddelay(10);
                msp_writereg(dvb, 0x12, 0x0013, 0x0c00);
                msp_writereg(dvb, 0x12, 0x0000, 0x7f00); // loudspeaker + headphone
                msp_writereg(dvb, 0x12, 0x0008, 0x0220); // loudspeaker source
                msp_writereg(dvb, 0x12, 0x0004, 0x7f00); // loudspeaker volume
                msp_writereg(dvb, 0x12, 0x000a, 0x0220); // SCART 1 source
                msp_writereg(dvb, 0x12, 0x0007, 0x7f00); // SCART 1 volume
                msp_writereg(dvb, 0x12, 0x000d, 0x4800); // prescale SCART

        }
        if (dvb->dvbtype == DVB_C) {
                // switch DVB SCART on
                outcom(dvb, COMTYPE_AUDIODAC, MainSwitch, 1, 0);
                outcom(dvb, COMTYPE_AUDIODAC, ADSwitch, 1, 1);
                
                //setgpio(dvb, 1, GPIO_OUTHI); // RGB on, SCART pin 16
                //setgpio(dvb, 3, GPIO_OUTLO); // SCARTpin 8
        }
        return 0;
}

__init static int 
find_saas(void)
{
        struct dvb_struct *dvb;
        struct saa7146 *saa;
        int card_type;
        int i, num=0;

#ifdef FIRM_MOD
        dvb_dpram_addr=dvb_root_addr=NULL;
        dvb_dpram_len=dvb_root_len=0;
        firm_mod=request_module("dvb_firm");
        printk("adpram=%08x, aroot=%08x, ldpram=%08x, lroot=%08x\n",
               (u32)dvb_dpram_addr, (u32)dvb_root_addr,
               dvb_dpram_len, dvb_root_len);
        printk("adpram=%08x, aroot=%08x, ldpram=%08x, lroot=%08x\n",
               (u32)virt_to_phys(dvb_dpram_addr), (u32)virt_to_phys(dvb_root_addr),
               dvb_dpram_len, dvb_root_len);
        printk("dpram[0]=%08x\n", *(int *)dvb_dpram_addr);
        printk("root[0]=%08x\n", *(int *)dvb_root_addr);

#endif
        for (i=0; i<8; i++) {
                saa=saa7146_get_handle(i);
                if (!saa)
                        continue;
                card_type=get_card_type(saa->device);
                if (card_type<0)
                        continue;
                dvb=&dvbs[num];
                init_card(dvb, num, saa->i2c_adap, card_type);
                num++;
        }
        return num;
}

static void 
dvb_irq(struct saa7146 *saa, u32 isr, void *data) 
{
        dprintk("dvb_irq\n");
        if (isr&MASK_19)
                debiirq(saa, data);
        if (isr&MASK_03)
                gpioirq(saa, data);
        if (isr&MASK_10)
                vpeirq(saa, data);
        if (isr&MASK_07)
                fidbirq(saa, data);
}

static int 
dvb_command(struct saa7146* saa, void *p, unsigned int cmd, void *arg)
{
        switch(cmd) {
        case SAA7146_SUSPEND:
                printk("dvb_suspend()\n");
                break;
        case SAA7146_RESUME:
                printk("dvb_resume()\n");
                break;
        default:
                return -ENOIOCTLCMD;
        }
        return 0;
}

static int dvb_extnum=0;

static int 
dvb_attach(struct saa7146* adap, void** p)
{
        if (dvb_extnum==MAX_NUM_DVB)
                return -1;
	*(struct dvb_struct**)p = &dvbs[dvb_extnum++];
        return 0;
}

static int 
dvb_detach(struct saa7146* adap, void** p)
{
        if (dvb_extnum==0)
                return -1;
     
        saa7146_write(adap->mem, IER, 
                      saa7146_read(adap->mem, IER) & 
                      ~(MASK_19|MASK_03));

        dvb_extnum--;
	*(struct dvb_struct**)p = 0;
        return 0;
}

static void 
dvb_inc_use(struct saa7146* adap)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

static void 
dvb_dec_use(struct saa7146* adap)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

static struct saa7146_extension dvb_extension = {
	"dvb extension\0",
	MASK_07|MASK_10|MASK_19|MASK_03|MASK_27,
	dvb_irq,
	dvb_command,
	dvb_attach,
	dvb_detach,
	dvb_inc_use,
	dvb_dec_use
};	


static int __init 
init_dvb(void) 
{
	int result = 0;
	num_dvb = 0;

	dprintk("dvb: init_dvb\n");

	if( 0 != (result = saa7146_add_extension(&dvb_extension))) {
		printk("dvb: extension registration failed, module not inserted.\n");
		return result;
	}

        if (!(num_dvb=find_saas()))
        {
	  printk(KERN_ERR "dvb: no dvb(s) found!\n");
	  return -ENODEV;
	}

        dprintk("dvb: dvbdevice found\n");

        printk(KERN_ERR "dvb: %d dvb(s) found!\n", num_dvb);
        return 0;
}

static void __exit 
exit_dvb(void)
{
	int i;

	fprintk("dvb: ==> cleanup_module\n");

	/* release the saa7146s */
	for( i = 0; i < num_dvb; i++) {
                del_timer(&dvbs[i].mon_timer);
		dvbs[i].mon_rmmod=1;
		wake_up_interruptible(&dvbs[i].mon_wait);
                while (dvbs[i].mon_thread)
                        ddelay(1);

                SetVoltage(&dvbs[i],0,0);
                saa7146_write(dvbs[i].saa->mem, IER, 
                              saa7146_read(dvbs[i].saa->mem, IER) & 
                              ~(MASK_19 | MASK_03));
                saa7146_write(dvbs[i].saa->mem, ISR,(MASK_19 | MASK_03));
                
                free_ipack(&dvbs[i].ipack[0]);
                free_ipack(&dvbs[i].ipack[1]);
                vfree(dvbs[i].iobuf);

                dvb_unregister(&dvbs[i]);
                vid_unregister(&dvbs[i]);
        }
        if ( saa7146_del_extension(&dvb_extension)) 
                printk(KERN_ERR "dvb: extension deregistration failed.\n");

	printk(KERN_ERR "dvb: %d dvb(s) released.\n", num_dvb);
}

module_init(init_dvb);
module_exit(exit_dvb);


/*
 * Local variables:
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -8
 * c-argdecl-indent: 8
 * c-label-offset: -8
 * c-continued-statement-offset: 8
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
